from dataclasses import dataclass, field
from typing import Dict, List, Tuple

import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal.auto.operation import Parameters, Qubits, Results, Routine

from . import resonator_flux_dependence, utils


# TODO: implement cross-talk
@dataclass
class QubitFluxParameters(Parameters):
    """QubitFlux runcard inputs."""

    freq_width: int
    """Width for frequency sweep relative to the qubit frequency (Hz)."""
    freq_step: int
    """Frequency step for sweep (Hz)."""
    bias_width: float
    """Width for bias sweep (V)."""
    bias_step: float
    """Bias step for sweep (V)."""
    nshots: int
    """Number of shots."""
    relaxation_time: int
    """Relaxation time (ns)."""
    drive_amplitude: float
    """Drive pulse amplitude. Same for all qubits."""


@dataclass
class QubitFluxResults(Results):
    """QubitFlux outputs."""

    sweetspot: Dict[List[Tuple], str] = field(metadata=dict(update="sweetspot"))
    """Sweetspot for each qubit."""
    frequency: Dict[List[Tuple], str] = field(metadata=dict(update="drive_frequency"))
    """Drive frequency for each qubit."""
    fitted_parameters: Dict[List[Tuple], List]
    """Raw fitting output."""


class QubitFluxData(resonator_flux_dependence.ResonatorFluxData):
    """QubitFlux acquisition outputs."""


def _acquisition(
    params: QubitFluxParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> QubitFluxData:
    """Data acquisition for QubitFlux Experiment."""
    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(
            qubit, start=0, duration=2000
        )
        qd_pulses[qubit].amplitude = params.drive_amplitude
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    delta_frequency_range = np.arange(
        -params.freq_width // 2, params.freq_width // 2, params.freq_step
    )
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        pulses=[qd_pulses[qubit] for qubit in qubits],
    )

    delta_bias_range = np.arange(
        -params.bias_width / 2, params.bias_width / 2, params.bias_step
    )
    bias_sweeper = Sweeper(
        Parameter.bias, delta_bias_range, qubits=list(qubits.values())
    )
    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency and flux bias
    data = QubitFluxData()

    # repeat the experiment as many times as defined by software_averages
    results = platform.sweep(
        sequence,
        bias_sweeper,
        freq_sweeper,
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
    )

    # retrieve the results for every qubit
    for qubit in qubits:
        result = results[ro_pulses[qubit].serial]

        biases = np.repeat(delta_bias_range, len(delta_frequency_range))
        freqs = np.array(
            len(delta_bias_range)
            * list(delta_frequency_range + ro_pulses[qubit].frequency)
        ).flatten()
        # store the results
        r = {k: v.ravel() for k, v in result.raw.items()}
        r.update(
            {
                "frequency[Hz]": freqs,
                "bias[V]": biases,
                "qubit": len(freqs) * [qubit],
            }
        )
        data.add_data_from_dict(r)

    return data


def _fit(data: QubitFluxData) -> QubitFluxResults:
    """Post-processing for QubitFlux Experiment."""
    return QubitFluxResults({}, {}, {})


def _plot(data: QubitFluxData, fit: QubitFluxResults, qubit):
    """Plotting function for QubitFlux Experiment."""
    return utils.flux_dependence_plot(data, fit, qubit)


qubit_flux = Routine(_acquisition, _fit, _plot)
"""QubitFlux Routine object."""



from typing import Optional

import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import FluxPulse, PulseSequence, Rectangular
from qibolab.sweeper import Parameter, Sweeper

from qibocal import plots
from qibocal.data import Data, DataUnits
from qibocal.decorators import plot


@plot("Flux pulse timing", plots.flux_pulse_timing)
def flux_pulse_timing(
    platform: AbstractPlatform,
    qubits: list,
    flux_pulse_amplitude_start,
    flux_pulse_amplitude_end,
    flux_pulse_amplitude_step,
    flux_pulse_start_start,
    flux_pulse_start_end,
    flux_pulse_start_step,
    flux_pulse_duration,
    time_window,
    nshots=1024,
    relaxation_time=None,
    software_averages=1,
    points=10,
):
    r"""
    The flux pulse timing experiment aims to determine the relative time of flight between
    the drive lines and the flux lines, as well as the time required for the transient response to
    disappear when the flux pulse stops. This is later required for the correct scheduling of
    the flux pulses in the implementation of 2q gates.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (list): List of target qubits to perform the action
        flux_pulse_amplitude_start (int): Initial flux pulse amplitude
        flux_pulse_amplitude_end (int): Maximum flux pulse amplitude
        flux_pulse_amplitude_step (int): Scan range step for the flux pulse amplitude
        flux_pulse_start_start (int): The earliest start time of the flux pluse, in ns,
            relative to the start of the first pulse of the sequence Ry(pi/2)
        flux_pulse_start_end (int): The earliest start time of the flux pluse, in ns
        flux_pulse_start_step (int): Scan range step for the flux pulse start
        flux_pulse_duration (int): The duration of the flux pulse
        time_window (int): The time window in ns between the end of the first pulse of the sequence Ry(pi/2)
            and the beginning of the second pulse of the sequence Ry(pi/2)
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data with the following keys
            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **prob[dimensionless]**: Statistical relative frequency of measuring state |1>
            - **flux_pulse_amplitude[dimensionless]**: Flux pulse amplitude
            - **flux_pulse_start[ns]**:The absolut start time of the flux pulse, in ns
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A Data object with the following parameters
            - **flux_pulse_duration**: The duration of the flux pulse
            - **time_window**: The time window in ns between the end of the first pulse of the sequence Ry(pi/2)
                    and the beginning of the second pulse of the sequence Ry(pi/2)
            - **initial_RY90_pulses_finish**: The absolut finish time of the first pulse of the sequence Ry(pi/2)
            - **qubit**: The qubit being tested

    """

    # the purpose of this routine is to determine the time of flight of a flux pulse and the duration
    # of the transient at the end of the pulse

    # 1) from state |0> apply Ry(pi/2) to state |+>,
    # 2) apply a detunning flux pulse of fixed duration at various start times
    # 3) after a certain time window from the end of the initial Ry(pi/2) pulse,
    #    measure in the X axis: Ry(pi/2) - MZ

    #   MX = Ry(pi/2) - (flux) - Ry(pi/2)  - MZ
    # If the flux pulse of sufficient amplitude falls within the time window between the Ry(pi/2) pulses,
    # it detunes the qubit and results in a rotation around the Z axis r so that MX = Cos(r)

    # reload instrument settings from runcard
    platform.reload_settings()

    # define the sequences of pulses to be executed
    sequence = PulseSequence()

    initial_RY90_pulses = {}
    flux_pulses = {}
    RY90_pulses = {}
    MZ_ro_pulses = {}
    if flux_pulse_start_start < 0:
        initial_RY90_pulses_start = -flux_pulse_start_start
    else:
        initial_RY90_pulses_start = 0

    for qubit in qubits:
        # start at |+> by rotating Ry(pi/2)
        initial_RY90_pulses[qubit] = platform.create_RX90_pulse(
            qubit, start=initial_RY90_pulses_start, relative_phase=np.pi / 2
        )

        # wait time window

        # rotate around the Y asis Ry(pi/2) to meassure X component
        RY90_pulses[qubit] = platform.create_RX90_pulse(
            qubit,
            start=initial_RY90_pulses[qubit].finish + time_window,
            relative_phase=np.pi / 2,
        )

        # add ro pulse at the end of each sequence
        MZ_ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RY90_pulses[qubit].finish
        )

        # apply a detuning flux pulse around the time window
        flux_pulses[qubit] = FluxPulse(
            start=initial_RY90_pulses_start + flux_pulse_start_start,
            duration=flux_pulse_duration,
            amplitude=flux_pulse_amplitude_start,  # fix for each run
            shape=Rectangular(),
            # relative_phase=0,
            channel=qubit.flux.name,
            qubit=qubit.name,
        )

        # add pulses to the sequences
        sequence.add(
            initial_RY90_pulses[qubit],
            flux_pulses[qubit],
            RY90_pulses[qubit],
            MZ_ro_pulses[qubit],
        )

    # define the parameters to sweep and their range:
    # flux pulse amplitude
    # flux pulse duration

    flux_pulse_amplitude_range = np.arange(
        flux_pulse_amplitude_start, flux_pulse_amplitude_end, flux_pulse_amplitude_step
    )
    sweeper = Sweeper(
        Parameter.amplitude,
        flux_pulse_amplitude_range,
        pulses=[flux_pulses[qubit] for qubit in qubits],
    )
    flux_pulse_start_range = np.arange(
        flux_pulse_start_start, flux_pulse_start_end, flux_pulse_start_step
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include flux pulse duration, amplitude and the probability
    data = DataUnits(
        name=f"data",
        quantities={
            "flux_pulse_amplitude": "dimensionless",
            "flux_pulse_start": "ns",
            "prob": "dimensionless",
        },
        options=["qubit", "iteration"],
    )

    parameters = Data(
        name=f"parameters",
        quantities=[
            "flux_pulse_duration",
            "time_window",
            "initial_RY90_pulses_finish",
            "qubit",
        ],
    )
    for qubit in qubits:
        parameters.add(
            {
                "flux_pulse_duration": flux_pulse_duration,
                "time_window": time_window,
                "initial_RY90_pulses_finish": initial_RY90_pulses[qubit].finish,
                "qubit": qubit,
            }
        )
    yield parameters

    ndata = len(flux_pulse_amplitude_range)
    count = 0
    for iteration in range(software_averages):
        # sweep the parameters
        for start in flux_pulse_start_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
            for qubit in qubits:
                flux_pulses[qubit].start = initial_RY90_pulses_start + start
            # execute the pulse sequence
            results = platform.sweep(
                sequence,
                sweeper,
                nshots=nshots,
                relaxation_time=relaxation_time,
                average=False,
            )
            for qubit in qubits:
                qubit_res = results[MZ_ro_pulses[qubit].serial]
                r = {
                    "MSR[V]": qubit_res.measurement.mean(axis=0),
                    "i[V]": qubit_res.i.mean(axis=0),
                    "q[V]": qubit_res.q.mean(axis=0),
                    "phase[rad]": qubit_res.phase.mean(axis=0),
                    "prob[dimensionless]": qubit_res.shots.mean(axis=0),
                    "flux_pulse_amplitude[dimensionless]": flux_pulse_amplitude_range,
                    "flux_pulse_start[ns]": ndata * [start],
                    "qubit": ndata * [qubit],
                    "iteration": ndata * [iteration],
                }
                data.add_data_from_dict(r)

            count += 1
    # finally, save the remaining data
    yield data


@plot("cryoscope_raw", plots.cryoscope_raw)
@plot("cryoscope_dephasing_heatmap", plots.cryoscope_dephasing_heatmap)
@plot("cryoscope_fft_peak_fitting", plots.cryoscope_fft_peak_fitting)
@plot("cryoscope_fft", plots.cryoscope_fft)
@plot("cryoscope_phase", plots.cryoscope_phase)
@plot("cryoscope_phase_heatmap", plots.cryoscope_phase_heatmap)
@plot("cryoscope_phase_unwrapped", plots.cryoscope_phase_unwrapped)
@plot("cryoscope_phase_unwrapped_heatmap", plots.cryoscope_phase_unwrapped_heatmap)
@plot(
    "cryoscope_phase_amplitude_unwrapped_heatmap",
    plots.cryoscope_phase_amplitude_unwrapped_heatmap,
)
@plot("cryoscope_detuning_time", plots.cryoscope_detuning_time)
@plot("cryoscope_distorted_amplitude_time", plots.cryoscope_distorted_amplitude_time)
@plot(
    "cryoscope_reconstructed_amplitude_time",
    plots.cryoscope_reconstructed_amplitude_time,
)
def cryoscope(
    platform: AbstractPlatform,
    qubits: dict,
    flux_pulse_amplitude,
    flux_pulse_duration_start,
    flux_pulse_duration_end,
    flux_pulse_duration_step,
    flux_pulse_amplitude_start,
    flux_pulse_amplitude_end,
    flux_pulse_amplitude_step,
    delay_before_readout,
    flux_pulse_shapes: Optional[list] = None,
    nshots=1024,
    relaxation_time=None,
    software_averages=1,
    points=10,
):
    r"""
    The cryoscope is one of the experiments required to characterise 2-qubit gates. Its aim is to measure
    the distorsions suffered by a flux pulse on its way from the control instrument to the qubit, so that they can later
    be corrected. Correcting these distorsions allows to reliably control the detuning of a qubit to a desired
    point where it interacts with a second qubit.
    In order to meassure the distorsions, the control qubit (the one with the highest frequency) is first prepared
    on a |+> state, a flux pulse of varying duration and amplitude is applied, finally the phase accumulated around
    the Z axis is meassured.
    For a given flux pulse amplitude, the derivative of the phase with respect to the pulse duration gives the instant
    detunning.
    Although the shape of the flux pulse sent is square, the impedance of the line from the instrument to the qubit,
    distort the pulse, resulting in a detuning with transients at the beginngin and end of the pulse.
    For every flux pulse amplitude, the stable detuning (after excluding those transients) can be measured.
    Those points are then fitted and this non-linear relation is used to determine the instant amplitude of the distorted
    pulse, as it is seen by the qubit, from the instant detunings measured.
    Once the distorted pulse is reconstructed, a convolution can be calculated such that pre-applied to the flux pulse,
    counters the distorsions.
    The type of convolutions chose are those that can be implemented in real time by control instruments: FIR and IIR
    filters.
    One filter may not be enough to counter all distorsions, so the cryoscope experiment can be iterated multiple times
    to determine multiple filters.

    https://arxiv.org/abs/1907.04818

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (list): List of target qubits to perform the action
        flux_pulse_duration_start (int): Initial flux pulse duration
        flux_pulse_duration_end (int): Maximum flux pulse duration
        flux_pulse_duration_step (int): Scan range step for the flux pulse duration
        flux_pulse_amplitude_start (int): Initial flux pulse amplitude
        flux_pulse_amplitude_end (int): Maximum flux pulse amplitude
        flux_pulse_amplitude_step (int): Scan range step for the flux pulse amplitude
        delay_before_readout (int): A time delay in ns between the end of the flux pulse and the beginning of the measurement of the phase
        flux_pulse_shapes (dict(PulseShape)): A dictionary of qubit_ids: PulseShape objects to be used for each qubit flux pulse.
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data with the following keys
            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **prob[dimensionless]**: Statistical relative frequency of measuring state |1>
            - **flux_pulse_amplitude[dimensionless]**: Flux pulse amplitude
            - **flux_pulse_duration[ns]**: Flux pulse duration in ns
            - **component**: The component being measured [MX, MY]
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

    """

    # 1) from state |0> apply Ry(pi/2) to state |+>,
    # 2) apply a flux pulse of variable duration,
    # 3) wait for the transient of the flux pulse to disappear
    # 3) measure in the X and Y axis
    #   MX = Ry(pi/2) - (flux)(t) - wait - Ry(pi/2)  - MZ
    #   MY = Ry(pi/2) - (flux)(t) - wait - Rx(-pi/2) - MZ
    # The flux pulse detunes the qubit and results in a rotation around the Z axis = atan(MY/MX)

    # reload instrument settings from runcard
    platform.reload_settings()

    # define the sequences of pulses to be executed
    MX_seq = PulseSequence()
    MY_seq = PulseSequence()

    initial_RY90_pulses = {}
    flux_pulses = {}
    RX90_pulses = {}
    RY90_pulses = {}
    MZ_ro_pulses = {}
    for qubit in qubits:
        # start at |+> by rotating Ry(pi/2)
        initial_RY90_pulses[qubit] = platform.create_RX90_pulse(
            qubit, start=0, relative_phase=np.pi / 2
        )

        if flux_pulse_shapes and len(flux_pulse_shapes) == len(qubits):
            flux_pulse_shape = eval(flux_pulse_shapes[qubit])
        else:
            flux_pulse_shape = Rectangular()

        # apply a detuning flux pulse
        flux_pulses[qubit] = FluxPulse(
            start=initial_RY90_pulses[qubit].se_finish,
            duration=flux_pulse_duration_start,  # sweep to produce oscillations [up to 400ns] in steps od 1ns? or 4?
            amplitude=flux_pulse_amplitude,  # fix for each run
            shape=flux_pulse_shape,
            # relative_phase=0,
            channel=platform.qubits[qubit].flux.name,
            qubit=platform.qubits[qubit].name,
        )

        # wait delay_before_readout

        # rotate around the X asis Rx(-pi/2) to meassure Y component
        RX90_pulses[qubit] = platform.create_RX90_pulse(
            qubit,
            start=flux_pulses[qubit].finish + delay_before_readout,
            relative_phase=np.pi,
        )

        # rotate around the Y asis Ry(pi/2) to meassure X component
        RY90_pulses[qubit] = platform.create_RX90_pulse(
            qubit,
            start=flux_pulses[qubit].finish + delay_before_readout,
            relative_phase=np.pi / 2,
        )

        # add ro pulse at the end of each sequence
        MZ_ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX90_pulses[qubit].finish
        )

        # add pulses to the sequences
        MX_seq.add(
            initial_RY90_pulses[qubit],
            flux_pulses[qubit],
            RY90_pulses[qubit],
            MZ_ro_pulses[qubit],
        )
        MY_seq.add(
            initial_RY90_pulses[qubit],
            flux_pulses[qubit],
            RX90_pulses[qubit],
            MZ_ro_pulses[qubit],
        )

        # DEBUG: Plot Cryoscope Sequences
        # MX_seq.plot("MX_seq")
        # MY_seq.plot("MY_seq")

    MX_tag = "MX"
    MY_tag = "MY"

    # define the parameters to sweep and their range:
    # flux pulse amplitude
    # flux pulse duration

    flux_pulse_amplitude_range = np.arange(
        flux_pulse_amplitude_start, flux_pulse_amplitude_end, flux_pulse_amplitude_step
    )
    sweeper = Sweeper(
        Parameter.amplitude,
        flux_pulse_amplitude_range,
        pulses=[flux_pulses[qubit] for qubit in qubits],
    )
    flux_pulse_duration_range = np.arange(
        flux_pulse_duration_start, flux_pulse_duration_end, flux_pulse_duration_step
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include flux pulse duration, amplitude and the probability
    data = DataUnits(
        name=f"data",
        quantities={
            "flux_pulse_duration": "ns",
            "flux_pulse_amplitude": "dimensionless",
            "prob": "dimensionless",
        },
        options=["component", "qubit", "iteration"],
    )

    ndata = len(flux_pulse_amplitude_range)
    count = 0
    for iteration in range(software_averages):
        for duration in flux_pulse_duration_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
            for qubit in qubits:
                flux_pulses[qubit].duration = duration
            # execute the pulse sequences
            for sequence, tag in [(MX_seq, MX_tag), (MY_seq, MY_tag)]:
                results = platform.sweep(
                    sequence,
                    sweeper,
                    nshots=nshots,
                    relaxation_time=relaxation_time,
                    average=False,
                )
                for qubit in qubits:
                    qubit_res = results[MZ_ro_pulses[qubit].serial]

                    r = {
                        "MSR[V]": qubit_res.measurement.mean(axis=0),
                        "i[V]": qubit_res.i.mean(axis=0),
                        "q[V]": qubit_res.q.mean(axis=0),
                        "phase[rad]": qubit_res.phase.mean(axis=0),
                        "prob[dimensionless]": qubit_res.shots.mean(axis=0),
                        "flux_pulse_duration[ns]": ndata * [duration],
                        "flux_pulse_amplitude[dimensionless]": flux_pulse_amplitude_range,
                        "component": ndata * [tag],
                        "qubit": ndata * [qubit],
                        "iteration": ndata * [iteration],
                    }
                    data.add_data_from_dict(r)
            count += 1

    # finally, save the remaining data
    yield data



from dataclasses import dataclass
from math import atan2

import numpy as np
import numpy.typing as npt

from .utils import identity


def constructor(_hyperparams):
    r"""Return the model class.

    Args:
        _hyperparams: Model hyperparameters.
    """
    return QubitFit()


def hyperopt(_x_train, _y_train, _path):
    r"""Perform an hyperparameter optimization and return the hyperparameters.

    Args:
        x_train: Training inputs.
        y_train: Training outputs.
        path (path): Model save path.

    Returns:
        Dictionary with model's hyperparameters.
    """
    return {}


normalize = identity


@dataclass
class QubitFit:
    r"""This class deploys a qubit state classifier.

    Args:
        iq_mean0 (np.ndarray): Center of the ground state cloud in the I-Q plane.
        iq_mean1 (np.ndarray): Center of the excited state cloud in the I-Q plane.
        threshold (float): Classifier's threshold.
        angle (float): Rotational angle.
    """  # TODO: add references

    iq_mean0: np.ndarray = np.array([0.0, 0.0])
    iq_mean1: np.ndarray = np.array([0.0, 0.0])
    threshold: float = 0.0
    angle: float = 0.0
    fidelity: float = None
    assignment_fidelity: float = None

    def fit(self, iq_coordinates, states: list):
        r"""Evaluate the model's parameters given the
        `iq_coordinates` and their relative ``states`
        (reference: <https://arxiv.org/abs/1004.4323>).
        """
        nshots = len(iq_coordinates)
        iq_state1 = iq_coordinates[(states == 1)]
        iq_state0 = iq_coordinates[(states == 0)]
        self.iq_mean0 = np.mean(iq_state0, axis=0)
        self.iq_mean1 = np.mean(iq_state1, axis=0)

        # translate
        iq_coordinates_translated = self.translate(iq_coordinates)
        iq_state1_trans = self.translate(self.iq_mean1)
        self.angle = -1 * atan2(iq_state1_trans[1], iq_state1_trans[0])

        # rotate
        iq_coord_rot = self.rotate(iq_coordinates_translated)

        x_values_state0 = np.sort(iq_coord_rot[(states == 0)][:, 0])
        x_values_state1 = np.sort(iq_coord_rot[(states == 1)][:, 0])

        # evaluate threshold and fidelity
        x_values = iq_coord_rot[:, 0]
        x_values.sort()
        cum_distribution_state0 = _eval_cumulative(x_values, x_values_state0)
        cum_distribution_state1 = _eval_cumulative(x_values, x_values_state1)

        cum_distribution_diff = np.abs(
            np.array(cum_distribution_state1) - np.array(cum_distribution_state0)
        )

        max_index = np.argmax(cum_distribution_diff)
        self.threshold = x_values[max_index]
        errors_state1 = nshots - cum_distribution_state1[max_index]
        errors_state0 = cum_distribution_state0[max_index]
        self.fidelity = cum_distribution_diff[max_index] / nshots
        self.assignment_fidelity = 1 - (errors_state1 + errors_state0) / nshots / 2

    def rotate(self, v):
        c, s = np.cos(self.angle), np.sin(self.angle)
        rot = np.array([[c, -s], [s, c]])
        return v @ rot.T

    def translate(self, v):
        return v - self.iq_mean0

    def predict(self, inputs: npt.NDArray):
        r"""Classify the `inputs`.

        Returns:
            List of predictions.
        """
        translated = self.translate(inputs)
        rotated = self.rotate(translated)
        return (rotated[:, 0] > self.threshold).astype(int)


def _eval_cumulative(input_data, points):
    r"""Evaluates in data the cumulative distribution
    function of `points`.
    WARNING: `input_data` and `points` should be sorted data.
    """
    # data and points sorted
    prob = []
    app = 0

    for val in input_data:
        app += np.max(np.searchsorted(points[app::], val) - 1, 0)
        prob.append(app + 1)

    return np.array(prob) / len(points)



import importlib
import json
import logging
import pathlib
import time
from dataclasses import asdict, dataclass
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score

from . import data, plots

CLS_MODULES = [
    "linear_svm",
    "ada_boost",
    "gaussian_process",
    "naive_bayes",
    "nn",
    "qubit_fit",
    "random_forest",
    "rbf_svm",
    "qblox_fit",
]


HYPERFILE = "hyperpars.json"
PREDFILE = "predictions.npy"
BENCHTABFILE = "benchmarks.csv"

base_dir = pathlib.Path()


def import_classifiers(cls_names: List[str]):
    r"""Return the classification models.

    Args:
        cls_name (list[str]): List of models' names.
    """
    importing_func = lambda mod: importlib.import_module(
        ".." + mod, "qibocal.fitting.classifier.*"
    )
    classifiers = list(map(importing_func, cls_names))
    return classifiers


class Classifier:
    r"""Classs to define the different classifiers used in the benchmarking.

    Args:
        mod: Classsification model.
        base_dir (Path): Where to store the classification results.

    """

    def __init__(self, mod, base_dir: pathlib.Path) -> None:
        self.mod = mod
        self.base_dir = base_dir

    @property
    def name(self):
        r"""Model's name."""
        return self.mod.__name__.split(".")[-1]

    @property
    def hyperopt(self):
        r"""The function that performs the hyperparameters optimization."""
        return self.mod.hyperopt

    @property
    def normalize(self):
        r"""The function that adds a data normalisation
        stage before the model classification.
        """
        return self.mod.normalize

    @property
    def constructor(self):
        r"""The model builder."""
        return self.mod.constructor

    @property
    def fit(self):
        r"""The model's fitting function."""
        return self.mod.fit

    @property
    def savedir(self):
        r"""The saving path."""
        return self.base_dir / self.name

    @property
    def hyperfile(self):
        r"""The path where the hyperparameters are stored."""
        return self.savedir / HYPERFILE

    @classmethod
    def load_model(cls, name: str, base_dir: pathlib.Path):
        r"""
        Giving the classification name this method returns the respective classifier.
        Args:
            name (str): classifier's name.
            base_dir (path): Where to store the classification results.
        Returns:
            Classification model.
        """
        inst = cls(import_classifiers([name]), base_dir)
        hyperpars = inst.load_hyper()
        return inst.create_model(hyperpars)

    @classmethod
    def model_from_dir(cls, folder: pathlib.Path):
        name = folder.name
        base_dir = folder.parent
        return cls.load_model(name, base_dir)

    def dump_hyper(self, hyperpars):
        r"""Saves the hyperparameters"""
        self.hyperfile.write_text(json.dumps(hyperpars, default=str), encoding="utf-8")

    def load_hyper(self):
        r"""Loads the hyperparameters and returns them."""
        return json.loads(self.hyperfile.load_text(encoding="utf-8"))

    def create_model(self, hyperpars):
        r"""Returns a model with the normalization stage.

        Args:
            hyperpars: Model's hyperparameters.

        Returns:
            Classification model.
        """
        return self.normalize(self.constructor(hyperpars))


@dataclass
class BenchmarkResults:
    r"""Class that stores the models information."""
    accuracy: float
    testing_time: float
    training_time: float
    name: Optional[str] = None


def benchmarking(model, x_train, y_train, x_test, y_test, **fit_kwargs):
    r"""This function evaluates the model's performances.
    Args:
        model: Classification model with `fit` and `predict` methods.
        x_train: Training input.
        y_train: Training output.
        x_test: Test input.
        y_test: Test output.
        **fit_kwargs:  Arbitrary keyword arguments for the `fit` function.

    Returns:
        - results (BenchmarkResults): Stores the model's accuracy, the training and testing time.
        - y_pred: Model's predictions.
        - model: trained model.
        - fit_info: Stores training infos.
    """
    # Evaluate training time
    start = time.time()
    fit_info = model.fit(x_train, y_train, **fit_kwargs)
    stop = time.time()
    training_time = stop - start
    # Evaluate test time per element
    start = time.time()
    y_pred = model.predict(x_test)
    stop = time.time()
    test_time = (stop - start) / len(x_test)
    # Evaluate accuracy
    score = accuracy_score(y_test, np.round(y_pred))
    logging.info(f"Accuracy: {score}")
    results = BenchmarkResults(score, test_time, training_time)

    return results, y_pred, model, fit_info


def plot_history(history, save_dir: pathlib.Path):
    r"""Plots the neural network history
    and save it in `json` file.

    Args:
        history (keras.callbacks.History): History.
        save_dir (Path): Storing path.
    """
    history_dict = history.history
    epochs = len(history_dict["loss"])
    plt.figure(figsize=(14, 7))
    plt.plot(range(epochs), history_dict["loss"], label="loss")
    plt.plot(range(epochs), history_dict["accuracy"], label="accurancy")
    plt.plot(range(epochs), history_dict["val_loss"], label="val_loss")
    plt.plot(range(epochs), history_dict["val_accuracy"], label="val_accuracy")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    plt.savefig(save_dir / "NN_training.pdf")
    json.dump(history_dict, open(save_dir / "NN_history.json", "w"))


def train_qubit(
    data_path: pathlib.Path, base_dir: pathlib.Path, qubit, classifiers=None
):
    r"""Given a dataset in `data_path` with qubits' information, this function performs the benchmarking of some classifiers.
    Each model's prediction `y_pred` is saved in  `basedir/qubit{qubit}/{classifier name}/predictions.npy`.

    Args:
        data_path(path): Where the qubits' data are stored.
        base_dir (path): Where save the results.
        qubit (int): Qubit ID.
        classifiers (list | None, optional): List of classification models. It must be a subset of `CLS_MODULES`.

    Returns:
        - benchmarks_table (pd.DataFrame): Table with the following columns

            - **name**: model's name
            - **accuracy**: model's accuracy
            - **training_time**: training time in seconds
            - **testing_time**: testing time per item in seconds.

        - y_test (list): List of test outputs.

    Example:
        The code below shows how to benchmark the state classifiers with `train_qubit` function:

            .. code-block:: python

                import logging
                from pathlib import Path

                from qibocal.fitting.classifier import plots, run

                logging.basicConfig(level=logging.INFO)
                # Define the data path
                data_path = Path("calibrate_qubit_states/data.csv")
                # Define the save path
                base_dir = Path("results")
                base_dir.mkdir(exist_ok=True)
                # Define the list of classifiers (not mandatory)
                classifiers = ["ada_boost", "linear_svm"]
                for qubit in range(1, 5):
                    print(f"QUBIT: {qubit}")
                    qubit_dir = base_dir / f"qubit{qubit}"
                    table, y_test, x_test = run.train_qubit(data_path, base_dir, qubit)
                    run.dump_benchmarks_table(table, qubit_dir)
                    # Plots
                    plots.plot_table(table, qubit_dir)
                    plots.plot_conf_matr(y_test, qubit_dir)
    """
    nn_epochs = 200
    nn_val_split = 0.2
    qubit_dir = base_dir / f"qubit{qubit}"
    qubit_dir.mkdir(exist_ok=True)
    qubit_data = data.load_qubit(data_path, qubit)
    data.plot_qubit(qubit_data, qubit_dir)
    x_train, x_test, y_train, y_test = data.generate_models(qubit_data)
    models = []
    results_list = []
    names = []
    if classifiers is None:
        classifiers = CLS_MODULES

    classifiers = import_classifiers(classifiers)

    for mod in classifiers:
        classifier = Classifier(mod, qubit_dir)
        classifier.savedir.mkdir(exist_ok=True)
        logging.info(f"Classification model: {classifier.name}")
        hyperpars = classifier.hyperopt(x_train, y_train, classifier.savedir)

        classifier.dump_hyper(hyperpars)
        model = classifier.create_model(hyperpars)

        if classifier.name == "nn":
            results, y_pred, model, fit_info = benchmarking(
                model,
                x_train,
                y_train,
                x_test,
                y_test,
                epochs=nn_epochs,
                validation_split=nn_val_split,
            )
            plot_history(fit_info, classifier.savedir)
        else:
            results, y_pred, model, _ = benchmarking(
                model, x_train, y_train, x_test, y_test
            )

        models.append(model)  # save trained model
        results.name = classifier.name
        results_list.append(results)
        names.append(classifier.name)
        dump_preds(y_pred, classifier.savedir)

    benchmarks_table = pd.DataFrame([asdict(res) for res in results_list])
    plots.plot_models_results(x_train, x_test, y_test, qubit_dir, models, names)
    plots.plot_roc_curves(x_test, y_test, qubit_dir, models, names)

    return benchmarks_table, y_test, x_test


def dump_preds(y_pred, dir_path):
    r"""Dumps the predictions in `{dir_path}/predictions.npy`.

    Args:
        y_pred (list): Predictions.
        dir_path (path): Saving path.
    """
    np.save(dir_path / PREDFILE, y_pred)


def dump_benchmarks_table(table, dir_path):
    r"""Dumps the benchmark table in `{dir_path}/benchmarks.csv`.

    Args:
        table (pd.DataFrame): Predictions.
        dir_path (path): Saving path.
    """
    table.to_csv(dir_path / BENCHTABFILE)


def preds_from_file(dir_path):
    r"""Load the predictions from a file

    Args:
        dir_path (path): Where the file `predictions.npy` is.

    Returns:
        Predictions.
    """
    return np.load(dir_path / PREDFILE)


def table_from_file(dir_path):
    r"""Load and return the benchmark table from a file

    Args:
        dir_path (path): Where the file `benchmarks.csv` is.
    """
    return pd.read_csv(dir_path / BENCHTABFILE)



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import res_spectroscopy_flux_fit
from qibocal.plots import frequency_attenuation, frequency_bias_flux


def snr(signal, noise):
    """Signal to Noise Ratio to detect peaks and valleys."""
    return 20 * np.log(signal / noise)


def choose_freq(freq, span, resolution):
    """Choose a new frequency gaussianly distributed around initial one.

    Args:
        freq (float): frequency we sample around from.
        span (float): search space we sample from.
        resolution (int): number of points for search space resolution.

    Returns:
        freq+ff (float): new frequency sampled gaussianly around old value.

    """
    g = np.random.normal(0, span / 10, 1)
    f = np.linspace(-span / 2, span / 2, resolution)
    for ff in f:
        if g <= ff:
            break
    return freq + ff


def get_noise(background, platform, ro_pulse, qubit, sequence):
    """Measure the MSR for the background noise at different points and average the results.

    Args:
        background (list): frequencies where no feature should be found.
        platform ():
        ro_pulse (): Used in order to execute the pulse sequence with the right parameters in the right qubit.
        qubit (int): TODO: might be useful to make this parameters implicit and not given.
        sequence ():

    Returns:
        noise (float): Averaged MSR value for the different background frequencies.

    """
    noise = 0
    for b_freq in background:
        ro_pulse.frequency = b_freq
        msr = platform.execute_pulse_sequence(sequence)[ro_pulse.serial].msr.mean()
        noise += msr
    return noise / len(background)


def scan_level(
    best_f,
    best_msr,
    max_runs,
    thr,
    span,
    resolution,
    noise,
    platform,
    ro_pulse,
    qubit,
    sequence,
):
    """Scans for the feature by sampling a gaussian near the previously found best point. Decides if the freature
    is found by checking if the snr against background noise is above a threshold.

    Args:
        best_f (int): Best found frequency in previous scan. Used as starting point.
        best_msr (float): MSR found for the previous best frequency. Used to check if a better value is found.
        max_runs (int): Maximum amount of tries before stopping if feature is not found.
        thr (float): SNR value used as threshold for detection of the feature.
        span (int): Search space around previous best value where the next frequency is sampled.
        resolution (int): How many points are taken in the span.
        noise (float): MSR value for the background noise.
        platform (AbstractPlatform): Platform the experiment is executed on.
        ro_pulse (ReadoutPulse): Used in order to execute the pulse sequence with the right parameters in the right qubit.
        qubit (int): qubit coupled to the resonator that we are probing.
        sequence (PulseSequence):

    Returns:
        best_f (float): Best frequency found for the feature.
        best_msr (float): MSR found for the feature.
        phase (float): Phase found for the feature.
        i (float): In-phase signal
        q (float): Quadrature signal.

    """
    freq = best_f
    for _ in range(max_runs):
        ro_pulse.frequency = freq
        result = platform.execute_pulse_sequence(sequence)[ro_pulse.serial]
        msr = result.msr.mean()
        if abs(snr(msr, noise)) >= thr:
            msr1 = msr
            if platform.resonator_type == "3D":
                msr = -msr
                best_msr = -best_msr
            if msr < best_msr:
                best_f, best_msr = freq, msr1
                return (
                    best_f,
                    best_msr,
                    result.phase.mean(),
                    result.i.mean(),
                    result.q.mean(),
                )
        freq = choose_freq(best_f, span, resolution)
    return best_f, best_msr, result.phase, result.i, result.q


def scan_small(best_f, best_msr, span, resolution, platform, ro_pulse, qubit, sequence):
    """Small scan around the found feature to fine-tune the measurement up to given precision.

    Args:
        best_f (int): Best found frequency in previous scan. Used as starting point.
        best_msr (float): MSR found for the previous best frequency. Used to check if a better value is found.
        span (int): Search space around previous best value where the next frequency is sampled.
        resolution (int): How many points are taken in the span. Taken as 10 for the small scan.
        platform (AbstractPlatform): Platform the experiment is executed on.
        ro_pulse (ReadoutPulse): Used in order to execute the pulse sequence with the right parameters in the right qubit.
        qubit (int): qubit coupled to the resonator that we are probing.
        sequence (PulseSequence):

    Returns:
        best_f (float): Best frequency found for the feature.
        best_msr (float): MSR found for the feature.
        phase (float): Phase found for the feature.
        i (float): In-phase signal.
        q (float): Quadrature signal.

    """
    start_f = best_f
    scan = np.linspace(-span / 2, span / 2, resolution)
    for s in scan:
        freq = start_f + s
        ro_pulse.frequency = freq
        result = platform.execute_pulse_sequence(sequence)[ro_pulse.serial]
        msr = result.msr.mean()
        msr1 = msr
        if platform.resonator_type == "3D":
            msr = -msr
            best_msr = -best_msr
        if msr < best_msr:
            best_f, best_msr = freq, msr1
    return best_f, best_msr, result.phase.mean(), result.i.mean(), result.q.mean()


@plot("Frequency vs Attenuation", frequency_attenuation)
def resonator_punchout_sample(
    platform: AbstractPlatform,
    qubits: dict,
    min_att,
    max_att,
    step_att,
    max_runs,
    thr,
    spans,
    small_spans,
    resolution,
    points=10,
):
    """Use gaussian samples to extract the punchout of the resonator for different values of attenuation.

    Args:
        platform (AbstractPlatform): Platform the experiment is executed on.
        qubits (dict): Dict of target Qubit objects to perform the action
        min_att (int): minimum attenuation value where the experiment starts. Less attenuation -> more power.
        max_att (int): maximum attenuation reached in the scan.
        step_att (int): change in attenuation after every step.
        max_runs (int): Maximum amount of tries before stopping if feature is not found.
        thr (float): SNR value used as threshold for detection of the feature.
        spans (list): Different spans to search for the feature at different precisions.
        small_spans (list): Different spans for the small sweeps to fine-tune the feature.
        resolution (int): How many points are taken in the span for the scan_level() function.
        points (int): Number of points plotted at a time in qq-live.

    Returns:
        data (Data): Data file with information on the feature response at each attenuation point.

    """

    for qubit in qubits:
        data = DataUnits(
            name=f"data_q{qubit}", quantities={"frequency": "Hz", "attenuation": "dB"}
        )

        platform.reload_settings()
        sequence = PulseSequence()
        ro_pulse = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulse)

        resonator_frequency = qubits[qubit].readout_frequency

        attenuation_range = np.arange(min_att, max_att, step_att)
        best_f = resonator_frequency

        opt_att = 30
        opt_snr = 0
        for k, att in enumerate(attenuation_range):
            platform.set_attenuation(qubit, att)
            background = [best_f + 1e7, best_f - 1e7]
            noise = get_noise(background, platform, ro_pulse, qubit, sequence)
            best_msr = noise
            for span in spans:
                best_f, best_msr, phase, i, q = scan_level(
                    best_f,
                    best_msr,
                    max_runs,
                    thr,
                    span,
                    resolution,
                    noise,
                    platform,
                    ro_pulse,
                    qubit,
                    sequence,
                )
            for span in small_spans:
                best_f, best_msr, phase, i, q = scan_small(
                    best_f, best_msr, span, 11, platform, ro_pulse, qubit, sequence
                )
            results = {
                "MSR[V]": best_msr,
                "i[V]": i,
                "q[V]": q,
                "phase[rad]": phase,
                "frequency[Hz]": best_f,
                "attenuation[dB]": att,
            }
            data.add(results)
            if k % points == 0:
                yield data
            if att >= opt_att:
                if abs(snr(best_msr, noise)) > opt_snr:
                    opt_snr = abs(snr(best_msr, noise))
                    opt_att = att
                    opt_f = best_f
        data1 = DataUnits(
            name=f"results_q{qubit}",
            quantities={"snr": "dimensionless", "frequency": "Hz", "attenuation": "dB"},
        )
        f_err = len(str(int(small_spans[-1] / 10)))
        results = {
            "snr[dimensionless]": opt_snr,
            "frequency[Hz]": round(opt_f, -f_err),
            "attenuation[dB]": opt_att,
        }
        data1.add(results)
        yield data1
        yield data


@plot("Frequency vs Current", frequency_bias_flux)
def resonator_flux_sample(
    platform: AbstractPlatform,
    qubits: dict,
    bias_min,
    bias_max,
    bias_step,
    fluxlines,
    max_runs,
    thr,
    spans,
    small_spans,
    resolution,
    params_fit,
    points=10,
):
    """Use gaussian samples to extract the flux-frequency response of the resonator for different values of bias.

    Args:
        platform (AbstractPlatform): Platform the experiment is executed on.
        qubits (dict): Dict of target Qubit objects to perform the action
        bias_min (float): minimum bias value where the experiment starts.
        bias_max (float): maximum bias reached in the scan.
        bias_step (float): change in bias after every step.
        fluxlines (list): ids of the bias lines to use for the experiment. Use qubit if matching the qubit id.
        max_runs (int): Maximum amount of tries before stopping if feature is not found.
        thr (float): SNR value used as threshold for detection of the feature.
        spans (list): Different spans to search for the feature at different precisions.
        small_spans (list): Different spans for the small sweeps to fine-tune the feature.
        resolution (int): How many points are taken in the span for the scan_level() function.
        params_fit (dic): Dictionary of parameters for the fit. {qubit : [freq_rh, g, Ec, Ej], ... }.
                          freq_rh is the resonator frequency at high power and g in the readout coupling.
                          If Ec and Ej are missing, the fit is valid in the transmon limit and if they are indicated,
                          contains the next-order correction.
        points (int): Number of points plotted at a time in qq-live.

    Returns:
        data (Data): Data file with information on the feature response at each bias point.

    """

    for qubit in qubits:
        params_fit = params_fit[qubit]
        for fluxline in fluxlines:
            data = DataUnits(
                name=f"data_q{qubit}_f{fluxline}",
                quantities={"frequency": "Hz", "bias": "V"},
            )

            if fluxline == "qubit":
                fluxline = qubit

            platform.reload_settings()
            for f in fluxlines:
                platform.set_bias(qubit, 0)

            sequence = PulseSequence()
            ro_pulse = platform.create_qubit_readout_pulse(qubit, start=0)
            sequence.add(ro_pulse)

            resonator_frequency = qubits[qubit].readout_frequency
            qubit_biasing_bias = qubits[qubit].sweetspot
            platform.set_bias(qubit, qubit_biasing_bias)
            bias_range = np.arange(bias_min, bias_max, bias_step) + qubit_biasing_bias
            start = next(
                (
                    index
                    for index, curr in enumerate(bias_range)
                    if curr >= qubit_biasing_bias
                )
            )
            start_f = resonator_frequency
            background = [start_f + 1e7, start_f - 1e7]
            noise = get_noise(background, platform, ro_pulse, qubit, sequence)
            bias_range = np.concatenate((bias_range[start:], bias_range[:start][::-1]))
            index = len(bias_range[start:])
            best_f = start_f
            for k, curr in enumerate(bias_range):
                if k == index:
                    best_f = start_f
                best_msr = noise
                platform.set_bias(qubit, curr)
                for span in spans:
                    best_f, best_msr, phase, i, q = scan_level(
                        best_f,
                        best_msr,
                        max_runs,
                        thr,
                        span,
                        resolution,
                        noise,
                        platform,
                        ro_pulse,
                        qubit,
                        sequence,
                    )
                for span in small_spans:
                    best_f, best_msr, phase, i, q = scan_small(
                        best_f, best_msr, span, 11, platform, ro_pulse, qubit, sequence
                    )
                results = {
                    "MSR[V]": best_msr,
                    "i[V]": i,
                    "q[V]": q,
                    "phase[rad]": phase,
                    "frequency[Hz]": best_f,
                    "bias[V]": curr,
                }
                data.add(results)
                if k % points == 0:
                    yield data
            yield res_spectroscopy_flux_fit(
                data,
                x="bias[V]",
                y="frequency[Hz]",
                qubit=qubit,
                fluxline=fluxline,
                params_fit=params_fit,
            )
            yield data



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.config import log
from qibocal.plots.utils import get_color


def rabi_amplitude_fit(x, p0, p1, p2, p3):
    # A fit to Superconducting Qubit Rabi Oscillation
    #   Offset                       : p[0]
    #   Oscillation amplitude        : p[1]
    #   Period    T                  : 1/p[2]
    #   Phase                        : p[3]
    #   Arbitrary parameter T_2      : 1/p[4]
    return p0 + p1 * np.sin(2 * np.pi * x * p2 + p3)


def rabi_length_fit(x, p0, p1, p2, p3, p4):
    # A fit to Superconducting Qubit Rabi Oscillation
    #   Offset                       : p[0]
    #   Oscillation amplitude        : p[1]
    #   Period    T                  : 1/p[2]
    #   Phase                        : p[3]
    #   Arbitrary parameter T_2      : 1/p[4]
    return p0 + p1 * np.sin(2 * np.pi * x * p2 + p3) * np.exp(-x * p4)


def plot(data, fit, qubit):
    if data.__class__.__name__ == "RabiAmplitudeData":
        quantity = "amplitude"
        unit = "dimensionless"
        title = "Amplitude (dimensionless)"
        fitting = rabi_amplitude_fit
    elif data.__class__.__name__ == "RabiLengthData":
        quantity = "length"
        unit = "ns"
        title = "Time (ns)"
        fitting = rabi_length_fit

    figures = []
    fitting_report = ""

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V)",
            "phase (rad)",
        ),
    )

    qubit_data = data.df[data.df["qubit"] == qubit]

    rabi_parameters = qubit_data[quantity].pint.to(unit).pint.magnitude.unique()
    fig.add_trace(
        go.Scatter(
            x=qubit_data[quantity].pint.to(unit).pint.magnitude,
            y=qubit_data["MSR"].pint.to("uV").pint.magnitude,
            marker_color=get_color(0),
            opacity=1,
            name="Voltage",
            showlegend=True,
            legendgroup="Voltage",
        ),
        row=1,
        col=1,
    )
    fig.add_trace(
        go.Scatter(
            x=qubit_data[quantity].pint.to(unit).pint.magnitude,
            y=qubit_data["phase"].pint.to("rad").pint.magnitude,
            marker_color=get_color(1),
            opacity=1,
            name="Phase",
            showlegend=True,
            legendgroup="Phase",
        ),
        row=1,
        col=2,
    )

    # add fitting trace
    if len(data) > 0:
        rabi_parameter_range = np.linspace(
            min(rabi_parameters),
            max(rabi_parameters),
            2 * len(data),
        )
        params = fit.fitted_parameters[qubit]
        fig.add_trace(
            go.Scatter(
                x=rabi_parameter_range,
                y=fitting(rabi_parameter_range, *params),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
                marker_color="rgb(255, 130, 67)",
            ),
            row=1,
            col=1,
        )

        fitting_report += (
            f"{qubit} | pi_pulse_amplitude: {float(fit.amplitude[qubit]):.3f}<br>"
        )
        fitting_report += (
            f"{qubit} | pi_pulse_length: {float(fit.length[qubit]):.3f}<br>"
        )

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title=title,
        yaxis_title="MSR (uV)",
        xaxis2_title=title,
        yaxis2_title="Phase (rad)",
    )

    figures.append(fig)

    return figures, fitting_report



"""Decorators implementation."""


def plot(header, method):
    """Decorator for adding plots in the report and live plotting page.

    Args:
        header (str): Header of the plot to use in the report.
        method (Callable): Plotting method defined under ``qibocal.plots``.
    """

    def wrapped(f):
        if hasattr(f, "plots"):
            # insert in the beginning of the list to have
            # proper plot ordering in the report
            f.plots.insert(0, (header, method))
        else:
            f.plots = [(header, method)]
        return f

    return wrapped



# TODO simulfilteredrb

from __future__ import annotations

from collections.abc import Iterable
from itertools import product

import numpy as np
import pandas as pd
from plotly.graph_objects import Figure
from qibo.models import Circuit
from qibo.noise import NoiseModel

import qibocal.calibrations.niGSC.basics.fitting as fitting_methods
from qibocal.calibrations.niGSC.basics.circuitfactory import SingleCliffordsFactory
from qibocal.calibrations.niGSC.basics.experiment import Experiment
from qibocal.calibrations.niGSC.basics.plot import Report, scatter_fit_fig
from qibocal.calibrations.niGSC.basics.utils import number_to_str


class ModuleFactory(SingleCliffordsFactory):
    pass


class ModuleExperiment(Experiment):
    """Inherits from abstract ``Experiment`` class."""

    def __init__(
        self,
        circuitfactory: Iterable,
        data: Iterable | None = None,
        nshots: int | None = None,
        noise_model: NoiseModel = None,
    ) -> None:
        """Calles the parent method and additionally prebuilds the circuit factory
        making it a list stored in memory and saves if ``save()`` method is called.

        Args:
            circuitfactory (Iterable): Gives a certain amount of circuits when
                iterated over.
            nshots (int): For execution of circuit, indicates how many shots.
            data (Iterable): If filled, ``data`` can be used to specifying parameters
                     while executing a circuit or deciding how to process results.
                     It is used to store all relevant data.
        """

        super().__init__(circuitfactory, data, nshots, noise_model)
        # Make the circuitfactory a list. That way they will be stored when
        # calling the save method and the circuits are not lost once executed.
        self.prebuild()
        self.name = "CorrelatedRB"

    def execute(self, circuit: Circuit, datarow: dict) -> dict:
        """Overwrited parents method. Executes a circuit, returns the single shot results and depth.

        Args:
            circuit (Circuit): Will be executed, has to return samples.
            datarow (dict): Dictionary with parameters for execution and
                immediate postprocessing information.
        """

        datadict = super().execute(circuit, datarow)
        # Measurement gate should not contribute to depth, therefore -1.
        # Take the amount of qubits into account.
        datadict["depth"] = int((circuit.ngates - 1) / len(datadict["samples"][0]))
        return datadict


class ModuleReport(Report):
    def __init__(self) -> None:
        super().__init__()
        self.title = "Correlated Filtered Randomized Benchmarking"


def filter_function(circuit: Circuit, datarow: dict) -> dict:
    """Calculates the filtered signal for every crosstalk irrep.

    Every irrep has a projector charactarized with a bit string
    :math:`\\boldsymbol{\\lambda}\\in\\mathbb{F}_2^N` where :math:`N` is the
    number of qubits.
    The experimental outcome for each qubit is denoted as
    :math:`\\ket{i_k}` with :math:`i_k=0, 1` with :math:`d=2`.

    .. math::
        f_{\\boldsymbol{\\lambda}}(i,g)
        = \\frac{1}{2^{N-|\\boldsymbol{\\lambda}|}}
            \\sum_{\\mathbf b\\in\\mathbb F_2^N}
            (-1)^{|\\boldsymbol{\\lambda}\\wedge\\mathbf b|}
            \\frac{1}{d^N}\\left(\\prod_{k=1}^N(d|\\bra{i_k} U_{g_{(k)}}
            \\ket{0}|^2)^{\\lambda_k-\\lambda_kb_k}\\right)

    Args:
        circuit (Circuit): The circuit used to produce the samples in ``datarow``.
        datarow (dict): Dictionary with samples produced by given ``circuit``.

    Returns:
        datarow (dict):  Filtered signals are stored additionally.
    """

    # Extract amount of used qubits and used shots.
    nshots, nqubits = datarow["samples"].shape
    # For qubits the local dimension is 2.
    d = 2
    # Fuse the gates for each qubit.
    fused_circuit = circuit.fuse(max_qubits=1)
    # Extract for each qubit the ideal state.
    # If depth = 0 there is only a measurement circuit and it does
    # not have an implemented matrix. Set the ideal states to ground states.
    if circuit.depth == 1:
        ideal_states = np.tile(np.array([1, 0]), nqubits).reshape(nqubits, 2)
    else:
        ideal_states = np.array(
            [fused_circuit.queue[k].matrix[:, 0] for k in range(nqubits)]
        )
    # Go through every irrep.
    f_list = []
    for l in np.array(list(product([False, True], repeat=nqubits))):
        # Check if the trivial irrep is calculated
        if not sum(l):
            # In the end every value will be divided by ``nshots``.
            a = nshots
        else:
            # Get the supported ideal outcomes and samples
            # for this irreps projector.
            suppl = ideal_states[l]
            suppsamples = datarow["samples"][:, l]
            a = 0
            # Go through all ``nshots`` samples
            for s in suppsamples:
                # Go through all combinations of (0,1) on the support
                # of lambda ``l``.
                for b in np.array(list(product([False, True], repeat=sum(l)))):
                    # Calculate the sign depending on how many times the
                    # nontrivial projector was used.
                    # Take the product of all probabilities chosen by the
                    # experimental outcome which are supported by the
                    # inverse of b.
                    a += (-1) ** sum(b) * np.prod(
                        d * np.abs(suppl[~b][np.eye(2, dtype=bool)[s[~b]]]) ** 2
                    )
        # Normalize with inverse of effective measuremetn.
        f_list.append(a * (d + 1) ** sum(l) / d**nqubits)
    for kk in range(len(f_list)):
        datarow[f"irrep{kk}"] = f_list[kk] / nshots
    return datarow


def post_processing_sequential(experiment: Experiment):
    """Perform sequential tasks needed to analyze the experiment results.

    The data is added/changed in the experiment, nothign has to be returned.

    Args:
        experiment (Experiment): Experiment object after execution of the experiment itself.
    """

    # Compute and add the ground state probabilities row by row.
    experiment.perform(filter_function)


def get_aggregational_data(experiment: Experiment) -> pd.DataFrame:
    """Computes aggregational tasks, fits data and stores the results in a data frame.

    No data is manipulated in the ``experiment`` object.

    Args:
        experiment (Experiment): After sequential postprocessing of the experiment data.

    Returns:
        pd.DataFrame: The summarized data.
    """

    # Needed for the amount of plots in the report
    nqubits = len(experiment.data[0]["samples"][0])
    data_list, index = [], []
    # Go through every irreducible representation projector used in the filter function.
    for kk in range(2**nqubits):
        # This has to match the label chosen in ``filter_function``.
        ylabel = f"irrep{kk}"
        depths, ydata = experiment.extract(ylabel, "depth", "mean")
        _, ydata_std = experiment.extract(ylabel, "depth", "std")
        # Fit an exponential without linear offset.
        popt, perr = fitting_methods.fit_exp1_func(depths, ydata)
        data_list.append(
            {
                "depth": depths,  # The x-axis.
                "data": ydata,  # The mean of ground state probability for each depth.
                "2sigma": 2 * ydata_std,  # The standard deviation error for each depth.
                "fit_func": "exp1_func",  # Which function was used to fit.
                "popt": {"A": popt[0], "p": popt[1]},  # The fitting paramters.
                "perr": {"A_err": perr[0], "p_err": perr[1]},  # The estimated errors.
            }
        )
        # Store the name to set is as row name for the data.
        index.append(ylabel)
    # Create a data frame out of the list with dictionaries.
    df = pd.DataFrame(data_list, index=index)
    return df


def build_report(experiment: Experiment, df_aggr: pd.DataFrame) -> Figure:
    """Use data and information from ``experiment`` and the aggregated data dataframe to
    build a reprot as plotly figure.

    Args:
        experiment (Experiment): After sequential postprocessing of the experiment data.
        df_aggr (pd.DataFrame): Normally build with ``get_aggregational_data`` function.

    Returns:
        (tuple): A `plotly.graphical_object.Figure` object and a string corresponding to a table.
    """

    # Initiate a report object.
    report = ModuleReport()
    # Add general information to the table.

    report.info_dict["Number of qubits"] = len(experiment.data[0]["samples"][0])
    report.info_dict["Number of shots"] = len(experiment.data[0]["samples"])
    report.info_dict["runs"] = experiment.extract("samples", "depth", "count")[1][0]
    lambdas = iter(product([0, 1], repeat=int(report.info_dict["Number of qubits"])))
    for kk, l in enumerate(lambdas):
        # Add the fitting parameters and  errors.
        dfrow = df_aggr.loc[f"irrep{kk}"]
        popt_pairs = (
            list(dfrow["popt"].items())[::2] + list(dfrow["popt"].items())[1::2]
        )
        report.info_dict[f"Irrep {l} Fit"] = "".join(
            [f"{key}={number_to_str(value)} " for key, value in popt_pairs]
        )
        perr_pairs = (
            list(dfrow["perr"].items())[::2] + list(dfrow["perr"].items())[1::2]
        )
        report.info_dict[f"Irrep {l} Deviations"] = "".join(
            [f"{key}={number_to_str(value)} " for key, value in perr_pairs]
        )
        # Use the predefined ``scatter_fit_fig`` function from ``basics.utils`` to build the wanted
        # plotly figure with the scattered filter function points and then mean per depth.
        figdict = scatter_fit_fig(experiment, df_aggr, "depth", f"irrep{kk}")
        # Add a subplot title for each irrep.
        figdict["subplot_title"] = f"Irrep {l}"
        report.all_figures.append(figdict)
    # Return the figure of the report object and the corresponding table.
    return report.build()



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.config import log
from qibocal.data import DataUnits

from . import utils


@dataclass
class ResonatorPunchoutParameters(Parameters):
    """ "ResonatorPunchout runcard inputs."""

    freq_width: int
    """Width for frequency sweep relative  to the readout frequency (hz)."""
    freq_step: int
    """Frequency step for sweep (Hz)."""
    min_amp_factor: float
    """Minimum amplitude multiplicative factor."""
    max_amp_factor: float
    """Maximum amplitude multiplicative factor."""
    step_amp_factor: float
    """Step amplitude multiplicative factor."""
    nshots: int
    """Number of shots."""
    relaxation_time: int
    """Relaxation time (ns)."""


@dataclass
class ResonatorPunchoutResults(Results):
    """ResonatorPunchout outputs."""

    readout_frequency: Dict[List[Tuple], str] = field(
        metadata=dict(update="readout_frequency")
    )
    """Readout frequency for each qubit."""
    readout_amplitude: Dict[List[Tuple], str] = field(
        metadata=dict(update="readout_amplitude")
    )
    """Readout amplitude for each qubit."""
    bare_frequency: Optional[Dict[List[Tuple], str]] = field(
        metadata=dict(update="bare_resonator_frequency")
    )
    """Bare resonator frequency for each qubit."""


class ResonatorPunchoutData(DataUnits):
    """ResonatorPunchout data acquisition."""

    def __init__(self, resonator_type=None):
        super().__init__(
            "data",
            {"frequency": "Hz", "amplitude": "dimensionless"},
            options=["qubit"],
        )
        self._resonator_type = resonator_type

    @property
    def resonator_type(self):
        """Type of resonator (2D or 3D)."""
        return self._resonator_type


def _acquisition(
    params: ResonatorPunchoutParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> ResonatorPunchoutData:
    """Data acquisition for Punchout over amplitude."""
    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()

    ro_pulses = {}
    for qubit in qubits:
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    # resonator frequency
    delta_frequency_range = np.arange(
        -params.freq_width // 2, params.freq_width // 2, params.freq_step
    )
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        [ro_pulses[qubit] for qubit in qubits],
    )

    # amplitude
    amplitude_range = np.arange(
        params.min_amp_factor, params.max_amp_factor, params.step_amp_factor
    )
    amp_sweeper = Sweeper(
        Parameter.amplitude, amplitude_range, [ro_pulses[qubit] for qubit in qubits]
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency and attenuation
    data = ResonatorPunchoutData(platform.resonator_type)

    # repeat the experiment as many times as defined by software_averages
    amps = np.repeat(amplitude_range, len(delta_frequency_range))

    results = platform.sweep(
        sequence,
        amp_sweeper,
        freq_sweeper,
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
    )

    # retrieve the results for every qubit
    for qubit, ro_pulse in ro_pulses.items():
        # average msr, phase, i and q over the number of shots defined in the runcard
        result = results[ro_pulse.serial]
        # store the results
        freqs = np.array(
            len(amplitude_range) * list(delta_frequency_range + ro_pulse.frequency)
        ).flatten()
        r = {k: v.ravel() for k, v in result.raw.items()}
        r.update(
            {
                "frequency[Hz]": freqs,
                "amplitude[dimensionless]": amps,
                "qubit": len(freqs) * [qubit],
            }
        )
        data.add_data_from_dict(r)

    return data


def _fit(data: ResonatorPunchoutData, fit_type="amplitude") -> ResonatorPunchoutResults:
    """Fit frequency and attenuation at high and low power for a given resonator."""

    qubits = data.df["qubit"].unique()

    bare_freqs = {}
    dressed_freqs = {}
    ro_amplitudes = {}

    for qubit in qubits:
        qubit_data = data.df[data.df["qubit"] == qubit].drop(
            columns=["i", "q", "qubit"]
        )

        try:
            normalised_data = qubit_data.groupby([fit_type], as_index=False)[
                ["MSR"]
            ].transform(utils.norm)

            averaged_data_updated = qubit_data.copy()
            averaged_data_updated.update(normalised_data["MSR"])

            min_points = utils.find_min_msr(
                averaged_data_updated, data.resonator_type, fit_type
            )
            vfunc = np.vectorize(lambda x: x.magnitude)
            min_points = vfunc(min_points)

            max_x = np.amax(min_points[:, 0])
            min_x = np.amin(min_points[:, 0])
            middle_x = (max_x + min_x) / 2

            hp_points = min_points[min_points[:, 0] < middle_x]
            lp_points = min_points[min_points[:, 0] >= middle_x]

            freq_hp = utils.get_max_freq(hp_points)
            freq_lp = utils.get_max_freq(lp_points)

            point_hp_max, point_hp_min = utils.get_points_with_max_freq(
                min_points, freq_hp
            )
            point_lp_max, point_lp_min = utils.get_points_with_max_freq(
                min_points, freq_lp
            )

            freq_lp = point_lp_max[0]
            ro_amp = point_lp_max[1]
            freq_hp = point_hp_max[0]

        except:
            log.warning("resonator_punchout_fit: the fitting was not succesful")
            freq_lp = 0.0
            freq_hp = 0.0
            ro_amp = 0.0

        dressed_freqs[qubit] = freq_lp / 1e9
        bare_freqs[qubit] = freq_hp / 1e9
        ro_amplitudes[qubit] = ro_amp

    return ResonatorPunchoutResults(
        dressed_freqs,
        ro_amplitudes,
        bare_freqs,
    )


def _plot(data: ResonatorPunchoutData, fit: ResonatorPunchoutResults, qubit):
    """Plotting function for ResonatorPunchout."""
    figures = []
    fitting_report = ""

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
        subplot_titles=(
            "Normalised MSR",
            "phase (rad)",
        ),
    )

    qubit_data = data.df[data.df["qubit"] == qubit].drop(columns=["qubit"])

    normalised_data = qubit_data.groupby(["amplitude"], as_index=False)[
        ["MSR"]
    ].transform(utils.norm)

    fig.add_trace(
        go.Heatmap(
            x=qubit_data["frequency"].pint.to("Hz").pint.magnitude,
            y=qubit_data["amplitude"].pint.to("dimensionless").pint.magnitude,
            z=normalised_data["MSR"].pint.to("dimensionless").pint.magnitude,
            colorbar_x=0.46,
        ),
        row=1,
        col=1,
    )
    fig.update_xaxes(title_text=f"{qubit}: Frequency (Hz)", row=1, col=1)
    fig.update_yaxes(title_text="Amplitude", row=1, col=1)
    fig.add_trace(
        go.Heatmap(
            x=qubit_data["frequency"].pint.to("Hz").pint.magnitude,
            y=qubit_data["amplitude"].pint.to("dimensionless").pint.magnitude,
            z=qubit_data["phase"].pint.to("rad").pint.magnitude,
            colorbar_x=1.01,
        ),
        row=1,
        col=2,
    )

    fig.add_trace(
        go.Scatter(
            x=[
                fit.readout_frequency[qubit] * 1e9,
            ],
            y=[
                fit.readout_amplitude[qubit],
            ],
            mode="markers",
            marker=dict(
                size=8,
                color="gray",
                symbol="circle",
            ),
        )
    )
    title_text = ""
    title_text += f"{qubit} | Resonator Frequency at Low Power:  {fit.readout_frequency[qubit]*1e9:,.0f} Hz<br>"
    title_text += f"{qubit} | Resonator Frequency at High Power: {fit.bare_frequency[qubit]*1e9:,.0f} Hz<br>"
    title_text += f"{qubit} | Readout Amplitude at Low Power: {fit.readout_amplitude[qubit]:,.3f} <br>"

    fitting_report = fitting_report + title_text

    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
    )

    figures.append(fig)

    return figures, fitting_report


resonator_punchout = Routine(_acquisition, _fit, _plot)
"""ResonatorPunchout Routine object."""



import json
import pathlib

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.colors import ListedColormap
from sklearn.inspection import DecisionBoundaryDisplay
from sklearn.metrics import RocCurveDisplay, confusion_matrix, roc_curve

from . import run

TABLEFILE = "benchmarks.pdf"
CONFMATRIXFIG = "confusion_matrices.pdf"
CONFMATRIXFILE = "confusion_matrices.json"
ROCFIG = "ROC_curves.pdf"
ROCFILE = "ROC_curves.json"
RESULTSFIG = "results.pdf"
COLOR = ListedColormap(["#FF0000", "#0000FF"])


def plot_table(table, path: pathlib.Path):
    r"""Plot the benchmark table and save it as
    `{path}/benchmarks.pdf`.

    Args:
        table (pd.DataFrame): Benchmark's table.
        path (path): Save path.
    """
    sns.set_style("darkgrid")
    g = sns.PairGrid(
        table,
        y_vars="name",
        x_vars=["accuracy", "testing_time", "training_time"],
        height=4,
        hue="name",
        palette="bright",
    )
    g.map(sns.scatterplot)
    plt.xscale("log")
    plt.tight_layout()
    plt.savefig(path / TABLEFILE)


def plot_conf_matr(y_test, path: pathlib.Path, classifiers=None):
    r"""
    Plot the confusion matrices of the `classifiers` and save it
    as  `{path}/confusion_matrices.pdf` and `{path}/confusion_matrices.json`.

    Args:
        y_test: Test outputs.
        path: Save path.
        classifiers (list | None, optional): List of classification models. It must be a subset of `run.CLS_MODULES`.
    """
    matrices = []
    names = []
    _figure = plt.figure(figsize=(30, 5))

    if classifiers is None:
        classifiers = run.CLS_MODULES

    classifiers = run.import_classifiers(classifiers)

    for count, model in enumerate(classifiers):
        classifier = run.Classifier(model, path)
        names.append(classifier.name)

        y_pred = np.load(path / classifier.name / run.PREDFILE)
        # Evaluate confusion matrices
        conf_matr = confusion_matrix(y_test, np.round(y_pred), normalize="true")
        matrices.append(conf_matr)
        # Plots
        ax = plt.subplot(1, len(classifiers), count + 1)
        sns.heatmap(
            conf_matr, annot=True, xticklabels=["P", "N"], yticklabels=["P", "N"]
        )
        ax.set_title(classifier.name)

    plt.tight_layout()
    plt.savefig(path / CONFMATRIXFIG)
    confusion_dic = {names[i]: matrices[i].tolist() for i in range(len(classifiers))}
    # Save data in JSON
    json.dump(
        confusion_dic,
        open(path / CONFMATRIXFILE, "w"),
        separators=(",", ":"),
        sort_keys=True,
        indent=4,
    )


def plot_roc_curves(x_test, y_test, path: pathlib.Path, models, models_names):
    r"""Plot the ROC curves of the `models` and save it
    as `{path}/ROC_curves.pdf` and `{path}/ROC_curves.json`.
    Args:
        x_test (list): Test inputs.
        y_test (list): Test outputs.
        path (path): Save path.
        models (list): List of trained classifiers.
        models_names (list[str]): List of classifiers' names.
    """
    _figure = plt.figure(figsize=(30, 5))
    fprs = []
    tprs = []

    len_list = len(models)
    for count, model in enumerate(models):
        ax = plt.subplot(1, len_list, count + 1)
        plt.subplot(1, len_list, count + 1)

        # Evaluate the ROC curves
        y_pred = np.load(path / models_names[count] / run.PREDFILE)
        fpr, tpr, _ = roc_curve(y_test, y_pred)
        fprs.append(fpr)
        tprs.append(tpr)

        try:
            RocCurveDisplay.from_estimator(
                model, x_test, y_test, ax=ax, color="darkorange"
            )
        except ValueError:
            RocCurveDisplay.from_predictions(
                y_test,
                y_pred,
                ax=ax,
                color="darkorange",
            )
        # Plot
        plt.plot([0, 1], [0, 1], "k--", label="chance level (AUC = 0.5)")
        plt.axis("square")
        plt.xlabel("False Positive Rate")
        plt.ylabel("True Positive Rate")
        plt.title(f"{models_names[count]}")
        plt.legend()
        plt.tight_layout()
        plt.savefig(path / ROCFIG)

    # Save data in a dictionary
    roc_dict = {
        models_names[i]: [tprs[i].tolist(), fprs[i].tolist()] for i in range(len_list)
    }
    json.dump(
        roc_dict,
        open(path / ROCFILE, "w"),
        separators=(",", ":"),
        sort_keys=True,
        indent=4,
    )


def plot_models_results(x_train, x_test, y_test, path, models, models_names):
    r"""Plot the decisions boundaries of the `c` and save it
    as `{path}/results.pdf`.
    Args:
        x_train (list): Train inputs.
        x_test (list): Test inputs.
        y_test (list): Test outputs.
        path (path): Save path.
        models (list): List of trained classifiers.
        models_names (list[str]): List of classifiers' names.
    """
    _figure = plt.figure(figsize=(20, 8))

    len_list = len(models)

    for count, classifier in enumerate(models):
        ax = plt.subplot(3, len_list // 3 + 1, count + 1)

        i, q = np.meshgrid(
            np.linspace(x_train[:, 0].min(), x_train[:, 0].max(), num=200),
            np.linspace(x_train[:, 1].min(), x_train[:, 1].max(), num=200),
        )
        grid = np.vstack([i.ravel(), q.ravel()]).T
        y_pred = np.reshape(classifier.predict(grid), q.shape)
        display = DecisionBoundaryDisplay(xx0=i, xx1=q, response=y_pred)

        display.plot(cmap="RdBu", alpha=0.8, ax=ax)
        cm_bright = COLOR
        ax.scatter(
            x_test[:, 0],
            x_test[:, 1],
            c=y_test,
            cmap=cm_bright,
            edgecolors="k",
        )

        ax.set_xticks(())
        ax.set_yticks(())
        ax.set_title(models_names[count])
        plt.tight_layout()
        plt.savefig(path / RESULTSFIG)



"""Track execution history."""
import copy
from dataclasses import dataclass
from typing import Dict, Tuple

from .operation import Results
from .runcard import Id
from .status import Status
from .task import Task


@dataclass
class Completed:
    """A completed task."""

    task: Task
    """A snapshot of the task when it was completed.

    .. todo::

        once tasks will be immutable, a separate `iteration` attribute should
        be added

    """
    res: Results
    """An object containing result parameters."""
    status: Status

    def __post_init__(self):
        self.task = copy.deepcopy(self.task)


class History(Dict[Tuple[Id, int], Completed]):
    """Execution history.

    This is not only used for logging and debugging, but it is an actual part
    of the execution itself, since later routines can retrieve the output of
    former ones from here.

    """

    def push(self, completed: Completed):
        self[(completed.task.id, completed.task.iteration)] = completed
        completed.task.iteration += 1

    # TODO: implemet time_travel()



"""Specify runcard layout, handles (de)serialization."""
from pathlib import Path
from typing import Any, Dict, List, NewType, Optional, Union

import yaml
from pydantic import Field
from pydantic.dataclasses import dataclass

from .operation import OperationId

Id = NewType("Id", str)
"""Action identifiers type."""


@dataclass
class Action:
    """Action specification in the runcard."""

    id: Id
    """Action unique identifier."""
    operation: Optional[OperationId] = None
    """Operation to be performed by the executor."""
    main: Optional[Id] = None
    """Main subsequent for action in normal flow."""
    next: Optional[Union[List[Id], Id]] = None
    """Alternative subsequent actions, branching from the current one."""
    priority: Optional[int] = None
    """Priority level, determining the execution order."""
    parameters: Optional[Dict[str, Any]] = None
    """Input parameters, either values or provider reference."""

    def __hash__(self) -> int:
        """Each action is uniquely identified by its id."""
        return hash(self.id)


@dataclass
class Runcard:
    """Structure of an execution runcard."""

    actions: List[Action]
    qubits: Optional[List[Union[int, str]]] = Field(default_factory=list)
    format: Optional[str] = None

    @classmethod
    def load(cls, card: Union[dict, Path]):
        """Load a runcard.

        It accepts both a dictionary, or a path to a YAML file, to be first
        deserialized in a dictionary, and further loaded into an instance.

        """
        content = (
            yaml.safe_load(card.read_text(encoding="utf-8"))
            if isinstance(card, Path)
            else card
        )
        return cls(**content)



from dataclasses import dataclass

import numpy as np
import plotly.graph_objects as go
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.data import Data
from qibocal.plots.utils import get_color

from . import allxy


@dataclass
class AllXYDragParameters(Parameters):
    """AllXYDrag runcard inputs."""

    beta_start: float
    """Initial beta parameter for Drag pulse."""
    beta_end: float
    """Final beta parameter for Drag pulse."""
    beta_step: float
    """Step beta parameter for Drag pulse."""


@dataclass
class AllXYDragResults(Results):
    """AllXYDrag outputs."""


class AllXYDragData(Data):
    """AllXYDrag acquisition outputs."""

    def __init__(self):
        super().__init__(
            name="data",
            quantities={
                "probability",
                "gateNumber",
                "beta_param",
                "qubit",
            },
        )


def _acquisition(
    params: AllXYDragParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> AllXYDragData:
    r"""
    Data acquisition for allXY experiment varying beta.
    The AllXY experiment is a simple test of the calibration of single qubit gatesThe qubit (initialized in the |0> state)
    is subjected to two back-to-back single-qubit gates and measured. In each round, we run 21 different gate pairs:
    ideally, the first 5 return the qubit to |0>, the next 12 drive it to superposition state, and the last 4 put the
    qubit in |1> state.

    The AllXY iteration method allows the user to execute iteratively the list of gates playing with the drag pulse shape
    in order to find the optimal drag pulse coefficient for pi pulses.
    """

    data = AllXYDragData()

    count = 0
    # sweep the parameters
    for beta_param in np.arange(
        params.beta_start, params.beta_end, params.beta_step
    ).round(4):
        gateNumber = 1
        for gates in allxy.gatelist:
            # create a sequence of pulses
            ro_pulses = {}
            sequence = PulseSequence()
            for qubit in qubits:
                sequence, ro_pulses[qubit] = allxy.add_gate_pair_pulses_to_sequence(
                    platform, gates, qubit, sequence, beta_param
                )

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            # retrieve the results for every qubit
            for ro_pulse in ro_pulses.values():
                z_proj = 2 * results[ro_pulse.serial].ground_state_probability - 1
                # store the results
                r = {
                    "probability": z_proj,
                    "gateNumber": gateNumber,
                    "beta_param": beta_param,
                    "qubit": ro_pulse.qubit,
                }
                data.add(r)
            gateNumber += 1
    return data


def _fit(_data: AllXYDragData) -> AllXYDragResults:
    """Post-processing for allXYDrag."""
    return AllXYDragResults()


def _plot(data: AllXYDragData, _fit: AllXYDragResults, qubit):
    """Plotting function for allXYDrag."""

    figures = []
    fitting_report = "No fitting data"

    fig = go.Figure()
    beta_params = data.df["beta_param"].unique()

    qubit_data = data.df[data.df["qubit"] == qubit]

    for j, beta_param in enumerate(beta_params):
        beta_param_data = qubit_data[qubit_data["beta_param"] == beta_param]
        fig.add_trace(
            go.Scatter(
                x=beta_param_data["gateNumber"],
                y=beta_param_data["probability"],
                marker_color=get_color(j),
                mode="markers+lines",
                opacity=0.5,
                name=f"Beta {beta_param}",
                showlegend=True,
                legendgroup=f"group{j}",
                text=allxy.gatelist,
                textposition="bottom center",
            ),
        )

    fig.add_hline(
        y=0,
        line_width=2,
        line_dash="dash",
        line_color="grey",
    )
    fig.add_hline(
        y=1,
        line_width=2,
        line_dash="dash",
        line_color="grey",
    )

    fig.add_hline(
        y=-1,
        line_width=2,
        line_dash="dash",
        line_color="grey",
    )

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Gate sequence number",
        yaxis_title="Expectation value of Z",
    )

    figures.append(fig)

    return figures, fitting_report


allxy_drag_pulse_tuning = Routine(_acquisition, _fit, _plot)
"""AllXYDrag Routine object."""



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal.auto.operation import Parameters, Qubits, Results, Routine

from .resonator_spectroscopy import ResonatorSpectroscopyData
from .utils import lorentzian_fit, spectroscopy_plot


@dataclass
class QubitSpectroscopyParameters(Parameters):
    """QubitSpectroscopy runcard inputs."""

    freq_width: int
    """Width for frequency sweep relative  to the qubit frequency."""
    freq_step: int
    """Frequency step for sweep."""
    drive_duration: int
    """Drive pulse duration. Same for all qubits."""
    drive_amplitude: Optional[float] = None
    """Drive pulse amplitude (optional). Same for all qubits."""
    nshots: int = 1024
    """Number of shots."""
    relaxation_time: int = 50
    """Relaxation time (ns)."""


@dataclass
class QubitSpectroscopyResults(Results):
    """QubitSpectroscopy outputs."""

    frequency: Dict[List[Tuple], str] = field(metadata=dict(update="drive_frequency"))
    """Drive frequecy for each qubit."""
    amplitude: Dict[List[Tuple], str]
    """Input drive amplitude. Same for all qubits."""
    fitted_parameters: Dict[List[Tuple], List]
    """Raw fitting output."""
    attenuation: Optional[Dict[List[Tuple], str]] = field(
        default_factory=dict,
    )
    """Input attenuation for each qubit (optional)."""


class QubitSpectroscopyData(ResonatorSpectroscopyData):
    """QubitSpectroscopy acquisition outputs."""


def _acquisition(
    params: QubitSpectroscopyParameters, platform: AbstractPlatform, qubits: Qubits
) -> QubitSpectroscopyData:
    """Data acquisition for qubit spectroscopy."""
    # create a sequence of pulses for the experiment:
    # long drive probing pulse - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(
            qubit, start=0, duration=params.drive_duration
        )
        if params.drive_amplitude is not None:
            qd_pulses[qubit].amplitude = params.drive_amplitude
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    delta_frequency_range = np.arange(
        -params.freq_width // 2, params.freq_width // 2, params.freq_step
    )
    sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        pulses=[qd_pulses[qubit] for qubit in qubits],
    )

    # Create data structure for data acquisition.
    data = QubitSpectroscopyData(
        platform.resonator_type, amplitude=params.drive_amplitude
    )

    results = platform.sweep(
        sequence,
        sweeper,
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
    )

    # retrieve the results for every qubit
    for qubit, ro_pulse in ro_pulses.items():
        # average msr, phase, i and q over the number of shots defined in the runcard
        result = results[ro_pulse.serial]
        r = result.raw
        # store the results
        r.update(
            {
                "frequency[Hz]": delta_frequency_range + qd_pulses[qubit].frequency,
                "qubit": len(delta_frequency_range) * [qubit],
            }
        )
        data.add_data_from_dict(r)
    return data


def _fit(data: QubitSpectroscopyData) -> QubitSpectroscopyResults:
    """Post-processing function for QubitSpectroscopy."""
    qubits = data.df["qubit"].unique()
    amplitudes = {}
    attenuations = {}
    frequency = {}
    fitted_parameters = {}
    for qubit in qubits:
        freq, fitted_params = lorentzian_fit(data, qubit)
        frequency[qubit] = freq
        amplitudes[qubit] = data.amplitude
        attenuations[qubit] = data.attenuation
        fitted_parameters[qubit] = fitted_params

    return QubitSpectroscopyResults(
        frequency=frequency,
        fitted_parameters=fitted_parameters,
        amplitude=amplitudes,
        attenuation=attenuations,
    )


def _plot(data: QubitSpectroscopyData, fit: QubitSpectroscopyResults, qubit):
    """Plotting function for QubitSpectroscopy."""
    return spectroscopy_plot(data, fit, qubit)


qubit_spectroscopy = Routine(_acquisition, _fit, _plot)
"""QubitSpectroscopy Routine object."""



import keras_tuner as kt  # pylint: disable=import-error
from tensorflow.keras import backend as K  # pylint: disable=import-error
from tensorflow.keras import callbacks, optimizers  # pylint: disable=import-error
from tensorflow.keras.layers import (  # pylint: disable=import-error
    BatchNormalization,
    Dense,
    Layer,
)
from tensorflow.keras.models import Sequential  # pylint: disable=import-error

from .utils import identity


def constructor(hyperpars):
    r"""Return the model class.

    Args:
        _hyperparams: Model hyperparameters.
    """
    best_hps = kt.HyperParameters.from_config(hyperpars)
    tuner = kt.Hyperband(_hypermodel)

    return tuner.hypermodel.build(best_hps)


def hyperopt(x_train, y_train, path):
    r"""Perform an hyperparameter optimization and return the hyperparameters.

    Args:
        x_train: Training inputs.
        y_train: Training outputs.
        path (path): Model save path.

    Returns:
        Dictionary with model's hyperparameters.
    """
    tuner = kt.Hyperband(
        _hypermodel,
        objective="val_accuracy",
        max_epochs=150,
        directory=path,
        project_name="NNmodel",
    )
    tuner.search_space_summary()

    stop_early = callbacks.EarlyStopping(monitor="val_accuracy", patience=10)

    tuner.search(
        x_train, y_train, epochs=120, validation_split=0.2, callbacks=[stop_early]
    )

    # Get the optimal hyperparameters
    best_hps = tuner.get_best_hyperparameters(num_trials=1)[0]

    return best_hps.get_config()


normalize = identity


class RBFLayer(Layer):
    r"""Deploy a neural network layer with RBF activation function.

    Args:
        units: Number of nodes in the layer.
        gamma: Activation function's parameter.
    """

    def __init__(self, units, gamma, **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.gamma = K.cast_to_floatx(gamma)

    def build(self, input_shape):
        self.mu = self.add_weight(
            name="mu",
            shape=(int(input_shape[1]), self.units),
            initializer="uniform",
            trainable=True,
        )
        super().build(input_shape)

    def call(self, inputs):
        diff = K.expand_dims(inputs) - self.mu
        l2 = K.sum(K.pow(diff, 2), axis=1)
        res = K.exp(-1 * self.gamma * l2)
        return res

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.units)


def _hypermodel(hp):
    r"""Hyper model for the Keras tuner"""
    hp_units_1 = hp.Int("units_1", min_value=16, max_value=1056, step=16)
    hp_units_2 = hp.Int("units_2", min_value=16, max_value=1056, step=16)
    activation = hp.Choice("activation", ["relu", "sigmoid", "tanh", "RBF"])
    learning_rate = hp.Float("learning_rate", min_value=1e-4, max_value=1e-2)
    optimizer_choice = hp.Choice("optimizer", ["Adam", "Adagrad", "SGD", "RMSprop"])
    losses = hp.Choice("losses", ["binary_crossentropy", "categorical_crossentropy"])

    model = Sequential()
    model.add(BatchNormalization())
    if activation == "RBF":
        model.add(
            RBFLayer(
                hp_units_1,
                hp.Float("mu", min_value=1e-4, max_value=1),
                input_shape=(2,),
            )
        )
    else:
        model.add(Dense(hp_units_1, input_shape=(2,), activation=activation))

    if activation == "RBF":
        model.add(
            RBFLayer(
                hp_units_2,
                hp.Float("mu", min_value=1e-4, max_value=1),
                input_shape=(2,),
            )
        )
    else:
        model.add(Dense(hp_units_2, input_shape=(2,), activation=activation))

    model.add(Dense(1, activation="sigmoid"))

    if optimizer_choice == "Adam":
        optimizer = optimizers.Adam(learning_rate=learning_rate)
    elif optimizer_choice == "Adagrad":
        optimizer = optimizers.Adagrad(learning_rate=learning_rate)
    elif optimizer_choice == "SGD":
        optimizer = optimizers.SGD(learning_rate=learning_rate)
    elif optimizer_choice == "RMSprop":
        optimizer = optimizers.RMSprop(learning_rate=learning_rate)
    else:
        raise ValueError

    model.compile(optimizer=optimizer, loss=losses, metrics=["accuracy"])
    return model



import numpy as np
from qibo.config import log
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import lorentzian_fit


@plot("MSR and Phase vs Resonator Frequency", plots.frequency_msr_phase)
def resonator_spectroscopy(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width: int,
    freq_step: int,
    nshots: int = 1024,
    relaxation_time: int = 50,
    software_averages: int = 1,
):
    r"""
    Perform spectroscopies on the qubits' readout resonators.
    This routine executes an initial scan around the expected resonator frequency indicated
    in the platform runcard. After that, a final sweep with more precision is executed centered in the new
    resonator frequency found.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): List of target qubits to perform the action
        freq_width (int): Width frequency in HZ to perform the high resolution sweep
        freq_step (int): Step frequency in HZ for the high resolution sweep
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - Two DataUnits objects with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Resonator frequency value in Hz
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **readout_frequency**: frequency
            - **peak_voltage**: peak voltage
            - **popt0**: Lorentzian's amplitude
            - **popt1**: Lorentzian's center
            - **popt2**: Lorentzian's sigma
            - **popt3**: Lorentzian's offset
            - **qubit**: The qubit being tested
    """
    # reload instrument settings from runcard
    platform.reload_settings()
    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    for qubit in qubits:
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range resonator frequency
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)
    sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        pulses=[ro_pulses[qubit] for qubit in qubits],
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency
    data = DataUnits(
        name="data",
        quantities={"frequency": "Hz"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    for iteration in range(software_averages):
        results = platform.sweep(
            sequence, sweeper, nshots=nshots, relaxation_time=relaxation_time
        )

        # retrieve the results for every qubit
        for qubit in qubits:
            # average msr, phase, i and q over the number of shots defined in the runcard
            result = results[ro_pulses[qubit].serial]
            # store the results
            r = result.raw
            r.update(
                {
                    "frequency[Hz]": delta_frequency_range + ro_pulses[qubit].frequency,
                    "qubit": len(delta_frequency_range) * [qubit],
                    "iteration": len(delta_frequency_range) * [iteration],
                }
            )
            data.add_data_from_dict(r)

        # finally, save the remaining data and fits
        yield data
        yield lorentzian_fit(
            data,
            x="frequency[GHz]",
            y="MSR[uV]",
            qubits=qubits,
            resonator_type=platform.resonator_type,
            labels=["readout_frequency", "peak_voltage"],
        )


@plot(
    "MSR and Phase vs Resonator Frequency and Attenuation",
    plots.frequency_attenuation_msr_phase,
)
@plot(
    "Cross section at half range attenuation", plots.frequency_attenuation_msr_phase_cut
)
def resonator_punchout_attenuation(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width,
    freq_step,
    min_att,
    max_att,
    step_att,
    relaxation_time=50,
    nshots=1024,
    software_averages=1,
):
    r"""
    Perform spectroscopies on the qubits' readout resonators, decreasing the attenuation applied to
    the read-out pulse, producing an increment of the power sent to the resonator.
    That shows the two regimes of a given resonator, low and high-power regimes.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): List of target qubits to perform the action
        freq_width (int): Width frequency in HZ to perform the spectroscopy sweep
        freq_step (int): Step frequency in HZ for the spectroscopy sweep
        min_att (int): Minimum value in db for the attenuation sweep
        max_att (int): Minimum value in db for the attenuation sweep
        step_att (int): Step attenuation in db for the attenuation sweep
        relaxation_time (int): Relaxation time between shots (ns)
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Resonator frequency value in Hz
            - **attenuation[dB]**: attenuation value in db applied to the flux line
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()

    ro_pulses = {}
    for qubit in qubits:
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    # resonator frequency
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        [ro_pulses[qubit] for qubit in qubits],
    )

    # attenuation
    attenuation_range = np.flip(np.arange(min_att, max_att, step_att))
    att_sweeper = Sweeper(Parameter.attenuation, attenuation_range, qubits=qubits)

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency and attenuation
    data = DataUnits(
        name=f"data",
        quantities={"frequency": "Hz", "attenuation": "dB"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    atts = np.repeat(attenuation_range, len(delta_frequency_range))
    for iteration in range(software_averages):
        results = platform.sweep(
            sequence,
            att_sweeper,
            freq_sweeper,
            nshots=nshots,
            relaxation_time=relaxation_time,
        )

        # retrieve the results for every qubit
        for qubit, ro_pulse in ro_pulses.items():
            # average msr, phase, i and q over the number of shots defined in the runcard
            result = results[ro_pulse.serial]
            # store the results
            freqs = np.array(
                len(attenuation_range)
                * list(delta_frequency_range + ro_pulse.frequency)
            ).flatten()
            r = result.raw
            r.update(
                {
                    "frequency[Hz]": freqs,
                    "attenuation[dB]": atts,
                    "qubit": len(freqs) * [qubit],
                    "iteration": len(freqs) * [iteration],
                }
            )
            data.add_data_from_dict(r)

        # save data
        yield data
        # TODO: calculate and save fit


@plot(
    "MSR and Phase vs Resonator Frequency and Amplitude",
    plots.frequency_amplitude_msr_phase,
)
def resonator_punchout(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width,
    freq_step,
    min_amp_factor,
    max_amp_factor,
    step_amp_factor,
    relaxation_time=None,
    nshots=1024,
    software_averages=1,
):
    r"""
    Perform spectroscopies on the qubits' readout resonators, decreasing the attenuation applied to
    the read-out pulse, producing an increment of the power sent to the resonator.
    That shows the two regimes of a given resonator, low and high-power regimes.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): List of target qubits to perform the action
        freq_width (int): Width frequency in HZ to perform the spectroscopy sweep
        freq_step (int): Step frequency in HZ for the spectroscopy sweep
        max_amp_factor (float): Maximum value of the factor that multiplies the amplitude
            of the readout pulse
        min_amp_factor (float): Minimum value of the factor that multiplies the amplitude
            of the readout pulse
        step_amp_factor (float): Step value of the factor that multiplies the amplitude
            of the readout pulse
        relaxation_time (int): Relaxation time between shots (ns)
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Resonator frequency value in Hz
            - **amplitude**: amplitude value of the pulse sent
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()

    ro_pulses = {}
    for qubit in qubits:
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    # resonator frequency
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        [ro_pulses[qubit] for qubit in qubits],
    )

    # amplitude
    amplitude_range = np.arange(min_amp_factor, max_amp_factor, step_amp_factor)
    amp_sweeper = Sweeper(
        Parameter.amplitude, amplitude_range, [ro_pulses[qubit] for qubit in qubits]
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency and attenuation
    data = DataUnits(
        name=f"data",
        quantities={"frequency": "Hz", "amplitude": "dimensionless"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    amps = np.repeat(amplitude_range, len(delta_frequency_range))
    for iteration in range(software_averages):
        results = platform.sweep(
            sequence,
            amp_sweeper,
            freq_sweeper,
            nshots=nshots,
            relaxation_time=relaxation_time,
        )

        # retrieve the results for every qubit
        for qubit, ro_pulse in ro_pulses.items():
            # average msr, phase, i and q over the number of shots defined in the runcard
            result = results[ro_pulse.serial]
            # store the results
            freqs = np.array(
                len(amplitude_range) * list(delta_frequency_range + ro_pulse.frequency)
            ).flatten()
            r = {k: v.ravel() for k, v in result.raw.items()}
            r.update(
                {
                    "frequency[Hz]": freqs,
                    "amplitude[dimensionless]": amps,
                    "qubit": len(freqs) * [qubit],
                    "iteration": len(freqs) * [iteration],
                }
            )
            data.add_data_from_dict(r)

        # save data
        yield data
        # TODO: calculate and save fit


@plot(
    "MSR and Phase vs Resonator Frequency and Flux",
    plots.frequency_flux_msr_phase,
)
def resonator_spectroscopy_flux(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width,
    freq_step,
    bias_width,
    bias_step,
    fluxlines,
    nshots=1024,
    relaxation_time=50,
    software_averages=1,
):
    r"""
    Perform spectroscopy on the readout resonator modifying the bias applied in the flux control line.
    This routine works for quantum devices flux controlled.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): List of target qubits to perform the action
        freq_width (int): Width frequency in HZ to perform the spectroscopy sweep
        freq_step (int): Step frequency in HZ for the spectroscopy sweep
        bias_width (float): Width bias in A for the flux bias sweep
        bias_step (float): Step bias in A for the flux bias sweep
        fluxlines (list): List of flux lines to use to perform the experiment. If it is set to "qubits", it uses each of
                        flux lines associated with the target qubits.
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Resonator frequency value in Hz
            - **bias[V]**: Current value in A applied to the flux line
            - **qubit**: The qubit being tested
            - **fluxline**: The fluxline being tested
            - **iteration**: The iteration number of the many determined by software_averages
    """
    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    for qubit in qubits:
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        [ro_pulses[qubit] for qubit in qubits],
    )

    # flux bias
    if fluxlines == "qubits":
        fluxlines = qubits

    delta_bias_range = np.arange(-bias_width / 2, bias_width / 2, bias_step)
    bias_sweeper = Sweeper(Parameter.bias, delta_bias_range, qubits=fluxlines)

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency and flux bias
    data = DataUnits(
        name=f"data",
        quantities={"frequency": "Hz", "bias": "V"},
        options=["qubit", "fluxline", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    for iteration in range(software_averages):
        results = platform.sweep(
            sequence,
            bias_sweeper,
            freq_sweeper,
            nshots=nshots,
            relaxation_time=relaxation_time,
        )

        # retrieve the results for every qubit
        for qubit, fluxline in zip(qubits, fluxlines):
            # TODO: Support more fluxlines for QM

            result = results[ro_pulses[qubit].serial]

            biases = np.repeat(
                delta_bias_range, len(delta_frequency_range)
            ) + platform.get_bias(fluxline)
            freqs = np.array(
                len(delta_bias_range)
                * list(delta_frequency_range + ro_pulses[qubit].frequency)
            ).flatten()
            # store the results
            r = {k: v.ravel() for k, v in result.raw.items()}
            r.update(
                {
                    "frequency[Hz]": freqs,
                    "bias[V]": biases,
                    "qubit": len(freqs) * [qubit],
                    "fluxline": len(freqs) * [fluxline],
                    "iteration": len(freqs) * [iteration],
                }
            )
            data.add_data_from_dict(r)

        # save data
        yield data


@plot("MSR and Phase vs Resonator Frequency", plots.dispersive_frequency_msr_phase)
def dispersive_shift(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width,
    freq_step,
    software_averages=1,
    points=10,
):
    r"""
    Perform spectroscopy on the readout resonator, with the qubit in ground and excited state, showing
    the resonator shift produced by the coupling between the resonator and the qubit.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): List of target qubits to perform the action
        freq_width (int): Width frequency in HZ to perform the spectroscopy sweep
        freq_step (int): Step frequency in HZ for the spectroscopy sweep
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained for the normal and shifted sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Resonator frequency value in Hz
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages
    """
    # TODO: add sweepers
    # reload instrument settings from runcard
    platform.reload_settings()

    # create 2 sequences of pulses for the experiment:
    # sequence_0: I  - MZ
    # sequence_1: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence_0 = PulseSequence()
    sequence_1 = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].duration
        )
        sequence_0.add(ro_pulses[qubit])
        sequence_1.add(qd_pulses[qubit])
        sequence_1.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)

    # create a DataUnits objects to store the results
    data_0 = DataUnits(
        name="data_0", quantities={"frequency": "Hz"}, options=["qubit", "iteration"]
    )
    data_1 = DataUnits(
        name="data_1", quantities={"frequency": "Hz"}, options=["qubit", "iteration"]
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    # TODO: implement sweeper
    for iteration in range(software_averages):
        # sweep the parameter
        for delta_freq in delta_frequency_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data_0
                yield data_1
                # calculate and save fit
                yield lorentzian_fit(
                    data=data_0,
                    x="frequency[Hz]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=["readout_frequency", "peak_voltage"],
                    fit_file_name="fit_data_0",
                )
                yield lorentzian_fit(
                    data=data_1,
                    x="frequency[Hz]",
                    y="MSR[uV]",
                    qubits=qubits,
                    resonator_type=platform.resonator_type,
                    labels=["readout_frequency_shifted", "peak_voltage"],
                    fit_file_name="fit_data_1",
                )

            # reconfigure the instruments based on the new resonator frequency
            # in this case setting the local oscillators
            # the pulse sequence does not need to be modified or recreated between executions
            for qubit in qubits:
                ro_pulses[qubit].frequency = (
                    delta_freq + qubits[qubit].readout_frequency
                )

            # execute the pulse sequences
            results_0 = platform.execute_pulse_sequence(sequence_0)
            results_1 = platform.execute_pulse_sequence(sequence_1)

            # retrieve the results for every qubit
            for data, results in list(zip([data_0, data_1], [results_0, results_1])):
                for ro_pulse in ro_pulses.values():
                    # average msr, phase, i and q over the number of shots defined in the runcard
                    r = results[ro_pulse.serial].average.raw
                    # store the results
                    r.update(
                        {
                            "frequency[Hz]": ro_pulse.frequency,
                            "qubit": ro_pulse.qubit,
                            "iteration": iteration,
                        }
                    )
                    data.add(r)
            count += 1
    # finally, save the remaining data and fits
    yield data_0
    yield data_1
    yield lorentzian_fit(
        data=data_0,
        x="frequency[Hz]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=["readout_frequency", "peak_voltage"],
        fit_file_name="fit_data_0",
    )
    yield lorentzian_fit(
        data=data_1,
        x="frequency[Hz]",
        y="MSR[uV]",
        qubits=qubits,
        resonator_type=platform.resonator_type,
        labels=["readout_frequency_shifted", "peak_voltage"],
        fit_file_name="fit_data_1",
    )



from qibocal.plots.allXY import *
from qibocal.plots.calibrate_qubit_states import *
from qibocal.plots.chevron import *
from qibocal.plots.cryoscope import *
from qibocal.plots.flipping import *
from qibocal.plots.rabi import *
from qibocal.plots.ramsey import *
from qibocal.plots.ro_optimization import *
from qibocal.plots.spectroscopies import *
from qibocal.plots.spin_echo import *
from qibocal.plots.t1 import *



import os.path

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.plots.utils import get_data_subfolders


def landscape_2q_gate(folder, routine, qubit, format):
    fitting_report = "No fitting data"
    highfreq = 2
    lowfreq = qubit
    if qubit > 2:
        highfreq = qubit
        lowfreq = 2

    subfolder = get_data_subfolders(folder)[0]
    data = DataUnits.load_data(
        folder, subfolder, routine, format, f"data_q{lowfreq}{highfreq}"
    )

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V) - High Frequency",
            "MSR (V) - Low Frequency",  # TODO: change this to <Z>
        ),
    )

    fig.add_trace(
        go.Scatter(
            x=data.get_values("theta", "rad")[data.df["q_freq"] == "high"][
                data.df["setup"] == "I"
            ].to_numpy(),
            y=data.get_values("MSR", "V")[data.df["q_freq"] == "high"][
                data.df["setup"] == "I"
            ].to_numpy(),
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Scatter(
            x=data.get_values("theta", "rad")[data.df["q_freq"] == "high"][
                data.df["setup"] == "X"
            ].to_numpy(),
            y=data.get_values("MSR", "V")[data.df["q_freq"] == "high"][
                data.df["setup"] == "X"
            ].to_numpy(),
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Scatter(
            x=data.get_values("theta", "rad")[data.df["q_freq"] == "low"][
                data.df["setup"] == "I"
            ].to_numpy(),
            y=data.get_values("MSR", "V")[data.df["q_freq"] == "low"][
                data.df["setup"] == "I"
            ].to_numpy(),
        ),
        row=1,
        col=2,
    )

    fig.add_trace(
        go.Scatter(
            x=data.get_values("theta", "rad")[data.df["q_freq"] == "low"][
                data.df["setup"] == "X"
            ].to_numpy(),
            y=data.get_values("MSR", "V")[data.df["q_freq"] == "low"][
                data.df["setup"] == "X"
            ].to_numpy(),
        ),
        row=1,
        col=2,
    )

    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="theta (rad)",
        yaxis_title="MSR (V)",
        xaxis2_title="theta (rad)",
        yaxis2_title="MSR (V)",
    )
    return [fig], fitting_report


def duration_amplitude_msr_flux_pulse(folder, routine, qubit, format):
    fitting_report = "No fitting data"
    highfreq = 2
    lowfreq = qubit
    if qubit > 2:
        highfreq = qubit
        lowfreq = 2

    subfolder = get_data_subfolders(folder)[0]
    data = DataUnits.load_data(
        folder, subfolder, routine, format, f"data_q{lowfreq}{highfreq}"
    )

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR (V) - High Frequency",
            "MSR (V) - Low Frequency",
        ),
    )

    fig.add_trace(
        go.Heatmap(
            x=data.get_values("flux_pulse_duration", "ns")[
                data.df["q_freq"] == "high"
            ].to_numpy(),
            y=data.get_values("flux_pulse_amplitude", "dimensionless")[
                data.df["q_freq"] == "high"
            ].to_numpy(),
            z=data.get_values("MSR", "V")[data.df["q_freq"] == "high"].to_numpy(),
            colorbar_x=0.45,
        ),
        row=1,
        col=1,
    )
    fig.add_trace(
        go.Heatmap(
            x=data.get_values("flux_pulse_duration", "ns")[
                data.df["q_freq"] == "low"
            ].to_numpy(),
            y=data.get_values("flux_pulse_amplitude", "dimensionless")[
                data.df["q_freq"] == "low"
            ].to_numpy(),
            z=data.get_values("MSR", "V")[data.df["q_freq"] == "low"].to_numpy(),
            colorbar_x=1.0,
        ),
        row=1,
        col=2,
    )
    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="duration (ns)",
        yaxis_title="amplitude (dimensionless)",
        xaxis2_title="duration (ns)",
        yaxis2_title="amplitude (dimensionless)",
    )
    return [fig], fitting_report


def duration_amplitude_I_flux_pulse(folder, routine, qubit, format):
    fitting_report = "No fitting data"
    highfreq = 2
    lowfreq = qubit
    if qubit > 2:
        highfreq = qubit
        lowfreq = 2

    subfolder = get_data_subfolders(folder)[0]
    data = DataUnits.load_data(
        folder, subfolder, routine, format, f"data_q{lowfreq}{highfreq}"
    )

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "I (V) - High Frequency",
            "I (V) - Low Frequency",
        ),
    )

    fig.add_trace(
        go.Heatmap(
            x=data.get_values("flux_pulse_duration", "ns")[
                data.df["q_freq"] == "high"
            ].to_numpy(),
            y=data.get_values("flux_pulse_amplitude", "dimensionless")[
                data.df["q_freq"] == "high"
            ].to_numpy(),
            z=data.get_values("i", "V")[data.df["q_freq"] == "high"].to_numpy(),
            colorbar_x=0.45,
        ),
        row=1,
        col=1,
    )
    fig.add_trace(
        go.Heatmap(
            x=data.get_values("flux_pulse_duration", "ns")[
                data.df["q_freq"] == "low"
            ].to_numpy(),
            y=data.get_values("flux_pulse_amplitude", "dimensionless")[
                data.df["q_freq"] == "low"
            ].to_numpy(),
            z=data.get_values("i", "V")[data.df["q_freq"] == "low"].to_numpy(),
            colorbar_x=1.0,
        ),
        row=1,
        col=2,
    )
    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="duration (ns)",
        yaxis_title="amplitude (dimensionless)",
        xaxis2_title="duration (ns)",
        yaxis2_title="amplitude (dimensionless)",
    )
    return [fig], fitting_report


def duration_amplitude_Q_flux_pulse(folder, routine, qubit, format):
    fitting_report = "No fitting data"
    highfreq = 2
    lowfreq = qubit
    if qubit > 2:
        highfreq = qubit
        lowfreq = 2

    subfolder = get_data_subfolders(folder)[0]
    data = DataUnits.load_data(
        folder, subfolder, routine, format, f"data_q{lowfreq}{highfreq}"
    )

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "Q (V) - High Frequency",
            "Q (V) - Low Frequency",
        ),
    )

    fig.add_trace(
        go.Heatmap(
            x=data.get_values("flux_pulse_duration", "ns")[
                data.df["q_freq"] == "high"
            ].to_numpy(),
            y=data.get_values("flux_pulse_amplitude", "dimensionless")[
                data.df["q_freq"] == "high"
            ].to_numpy(),
            z=data.get_values("q", "V")[data.df["q_freq"] == "high"].to_numpy(),
            colorbar_x=0.45,
        ),
        row=1,
        col=1,
    )
    fig.add_trace(
        go.Heatmap(
            x=data.get_values("flux_pulse_duration", "ns")[
                data.df["q_freq"] == "low"
            ].to_numpy(),
            y=data.get_values("flux_pulse_amplitude", "dimensionless")[
                data.df["q_freq"] == "low"
            ].to_numpy(),
            z=data.get_values("q", "V")[data.df["q_freq"] == "low"].to_numpy(),
            colorbar_x=1.0,
        ),
        row=1,
        col=2,
    )
    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="duration (ns)",
        yaxis_title="amplitude (dimensionless)",
        xaxis2_title="duration (ns)",
        yaxis2_title="amplitude (dimensionless)",
    )
    return [fig], fitting_report



from typing import Union

import numpy as np

# Gates, without having to define any paramters
ONEQ_GATES = ["I", "X", "Y", "Z", "H", "S", "SDG", "T", "TDG"]


def effective_depol(error_channel, **kwargs):
    """Computes the effective depolarizing error of a channel.

    Args:
        error_channel (qibo.gates.Channel): Noise channel with `to_pauli_liouville` representation.

    Returns:
        float: Effective depolarizing error of a given error_channel.
    """
    liouvillerep = error_channel.to_pauli_liouville(normalize=True)
    d = int(np.sqrt(len(liouvillerep)))
    depolp = (np.trace(liouvillerep) - 1) / (d**2 - 1)
    return depolp


def probabilities(allsamples: Union[list, np.ndarray]) -> np.ndarray:
    """Takes the given list/array (3-dimensional) of samples and returns probabilities
    for each possible state to occure.

    The states for 4 qubits are order as follows:
    [(0, 0, 0, 0), (0, 0, 0, 1), (0, 0, 1, 0), (0, 0, 1, 1), (0, 1, 0, 0),
    (0, 1, 0, 1), (0, 1, 1, 0), (0, 1, 1, 1), (1, 0, 0, 0), (1, 0, 0, 1),
    (1, 0, 1, 0), (1, 0, 1, 1), (1, 1, 0, 0), (1, 1, 0, 1), (1, 1, 1, 0), (1, 1, 1, 1)]

    Args:
        allsamples (Union[list, np.ndarray]): The single shot samples, 3-dimensional.

    Returns:
        np.ndarray: Probability array of 2 dimension.
    """

    from itertools import product

    # Make it an array to use the shape property.
    allsamples = np.array(allsamples)
    # The array has to have three dimension.
    if len(allsamples.shape) == 2:
        allsamples = allsamples[None, ...]
    nqubits, nshots = len(allsamples[0][0]), len(allsamples[0])
    # Create all possible state vectors.
    allstates = list(product([0, 1], repeat=nqubits))
    # Iterate over all the samples and count the different states.
    probs = [
        [np.sum(np.product(samples == state, axis=1)) for state in allstates]
        for samples in allsamples
    ]
    probs = np.array(probs) / (nshots)
    return probs


def gate_fidelity(eff_depol: float, primitive=False) -> float:
    """Returns the average gate fidelity given the effective
    depolarizing parameter for single qubits.

    If primitive is True, divide by additional 1.875 as convetion in RB reporting.
    (The original reasoning was that Clifford gates are typically
    compiled with an average number of 1.875 Pi half pulses.)

    Args:
        eff_depol (float): The effective depolarizing parameter.
        primitive (bool, optional): If True, additionally divide by 1.875.

    Returns:
        float: Average gate fidelity
    """
    infidelity = (1 - eff_depol) / 2
    if primitive:
        infidelity /= 1.875
    return 1 - infidelity


def number_to_str(number: Union[int, float, complex]) -> str:
    """Converts a number into a string.
    Returns only the real number if imaginary part is < 1e-3.

    Args:
        number (int | float | complex)

    Returns:
        str: The number expressed as a string, with three floating points.
    """
    return (
        f"{np.real(number):.3f}" if np.abs(np.imag(number)) < 1e-3 else f"{number:.3f}"
    )



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import ro_optimization_fit


@plot("Qubit States", plots.ro_frequency)
def ro_frequency(
    platform: AbstractPlatform,
    qubits: dict,
    frequency_width: float,
    frequency_step: float,
    nshots: int,
):
    """
    Method which optimizes the Read-out fidelity by varying the Read-out pulse frequency.
    Two analogous tests are performed for calibrate the ground state and the excited state of the oscillator.
    The subscripts `exc` and `gnd` will represent the excited state |1> and the ground state |0>.
    Their distinctiveness is then associated to the fidelity.

    Args:
        platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): custom abstract platform on which we perform the calibration.
        qubits (dict): List of target qubits to perform the action
        nshots (int): number of times the pulse sequence will be repeated.
        frequency_width (float): width of the frequency range to be swept in Hz.
        frequency_step (float): step of the frequency range to be swept in Hz.
    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **iteration**: Execution number
            - **qubit**: The qubit being tested
            - **state**: The state of the qubit being tested
            - **frequency[Hz]**: The frequency of the readout being tested
            - **delta_frequency[Hz]**: The frequency offset from the runcard value

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create two sequences of pulses:
    # state0_sequence: I  - MZ
    # state1_sequence: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    state0_sequence = PulseSequence()
    state1_sequence = PulseSequence()

    RX_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        RX_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX_pulses[qubit].finish
        )

        state0_sequence.add(ro_pulses[qubit])
        state1_sequence.add(RX_pulses[qubit])
        state1_sequence.add(ro_pulses[qubit])
    sequences = {0: state0_sequence, 1: state1_sequence}
    # create a DataUnits object to store the results
    data = DataUnits(
        name="data",
        quantities={"frequency": "Hz", "delta_frequency": "Hz"},
        options=["qubit", "iteration", "state"],
    )

    # iterate over the frequency range
    delta_frequency_range = np.arange(
        -frequency_width / 2, frequency_width / 2, frequency_step
    )

    frequency_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        pulses=[ro_pulses[qubit] for qubit in qubits],
    )

    # Execute sequences for both states
    for state in [0, 1]:
        results = platform.sweep(
            sequences[state], frequency_sweeper, nshots=nshots, average=False
        )

        # retrieve and store the results for every qubit)
        for qubit in qubits:
            r = {k: v.ravel() for k, v in results[ro_pulses[qubit].serial].raw.items()}
            r.update(
                {
                    "frequency[Hz]": np.repeat(
                        np.vstack(delta_frequency_range).T,
                        nshots,
                        axis=0,
                    ).flatten()
                    + ro_pulses[qubit].frequency,
                    "delta_frequency[Hz]": np.repeat(
                        np.vstack(delta_frequency_range).T,
                        nshots,
                        axis=0,
                    ).flatten(),
                    "qubit": [qubit] * nshots * len(delta_frequency_range),
                    "iteration": np.repeat(
                        np.vstack(np.arange(nshots)).T,
                        len(delta_frequency_range),
                        axis=1,
                    ).flatten(),
                    "state": [state] * nshots * len(delta_frequency_range),
                }
            )
            data.add_data_from_dict(r)

    # finally, save the remaining data and the fits
    yield data
    yield ro_optimization_fit(data, "state", "qubit", "iteration", "delta_frequency")


@plot("Qubit States", plots.ro_amplitude)
def ro_amplitude(
    platform: AbstractPlatform,
    qubits: dict,
    amplitude_factor_min: float,
    amplitude_factor_max: float,
    amplitude_factor_step: float,
    nshots: int,
):
    """
    Method which optimizes the Read-out fidelity by varying the Read-out pulse amplitude.
    Two analogous tests are performed for calibrate the ground state and the excited state of the oscillator.
    The subscripts `exc` and `gnd` will represent the excited state |1> and the ground state |0>.
    Their distinctiveness is then associated to the fidelity.

    Args:
        platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): custom abstract platform on which we perform the calibration.
        qubits (dict): List of target qubits to perform the action
        nshots (int): number of times the pulse sequence will be repeated.
        amplitude_factor_min (float): minimum amplitude factor to be swept.
        amplitude_factor_max (float): maximum amplitude factor to be swept.
        amplitude_factor_step (float): step of the amplitude factor to be swept.
    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **qubit**: The qubit being tested
            - **iteration**: Execution number
            - **state**: The state of the qubit being tested
            - **amplitude_factor**: The amplitude factor of the readout being tested
            - **delta_amplitude_factor**: The amplitude offset from the runcard value

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create two sequences of pulses:
    # state0_sequence: I  - MZ
    # state1_sequence: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    state0_sequence = PulseSequence()
    state1_sequence = PulseSequence()

    RX_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        RX_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX_pulses[qubit].finish
        )

        state0_sequence.add(ro_pulses[qubit])
        state1_sequence.add(RX_pulses[qubit])
        state1_sequence.add(ro_pulses[qubit])
    sequences = {0: state0_sequence, 1: state1_sequence}
    # create a DataUnits object to store the results
    data = DataUnits(
        name="data",
        quantities={"amplitude": "dimensionless", "delta_amplitude": "dimensionless"},
        options=["qubit", "iteration", "state"],
    )

    # iterate over the amplitude range
    delta_amplitude_range = np.arange(
        amplitude_factor_min, amplitude_factor_max, amplitude_factor_step
    )

    amplitude_sweeper = Sweeper(
        Parameter.amplitude,
        delta_amplitude_range,
        pulses=[ro_pulses[qubit] for qubit in qubits],
    )

    # Execute sequences for both states
    for state in [0, 1]:
        results = platform.sweep(
            sequences[state], amplitude_sweeper, nshots=nshots, average=False
        )

        # retrieve and store the results for every qubit)
        for qubit in qubits:
            r = {k: v.ravel() for k, v in results[ro_pulses[qubit].serial].raw.items()}
            r.update(
                {
                    "amplitude[dimensionless]": np.repeat(
                        np.vstack(delta_amplitude_range).T,
                        nshots,
                        axis=0,
                    ).flatten()
                    * ro_pulses[qubit].amplitude,
                    "delta_amplitude[dimensionless]": np.repeat(
                        np.vstack(delta_amplitude_range).T,
                        nshots,
                        axis=0,
                    ).flatten(),
                    "qubit": [qubit] * nshots * len(delta_amplitude_range),
                    "iteration": np.repeat(
                        np.vstack(np.arange(nshots)).T,
                        len(delta_amplitude_range),
                        axis=1,
                    ).flatten(),
                    "state": [state] * nshots * len(delta_amplitude_range),
                }
            )
            data.add_data_from_dict(r)

    # finally, save the remaining data and the fits
    yield data
    yield ro_optimization_fit(data, "state", "qubit", "iteration", "delta_amplitude")


@plot("TWPA frequency", plots.ro_frequency)
def twpa_frequency(
    platform: AbstractPlatform,
    qubits: dict,
    frequency_width: float,
    frequency_step: float,
    nshots: int,
):
    """
    Method which optimizes the Read-out fidelity by varying the frequency of the TWPA.
    Two analogous tests are performed for calibrate the ground state and the excited state of the oscillator.
    The subscripts `exc` and `gnd` will represent the excited state |1> and the ground state |0>.
    Their distinctiveness is then associated to the fidelity.

    Args:
        platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): custom abstract platform on which we perform the calibration.
        qubits (dict): List of target qubits to perform the action
        frequency_width (float): Frequency range to sweep in Hz
        frequency_step (float): Frequency step to sweep in Hz
        nshots (int): number of times the pulse sequence will be repeated.
    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **qubit**: The qubit being tested
            - **iteration**: Execution number
            - **state**: The state of the qubit being tested
            - **frequency**: The frequency of the TWPA being tested
            - **delta_frequency**: The frequency offset from the runcard value

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create two sequences of pulses:
    # state0_sequence: I  - MZ
    # state1_sequence: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    state0_sequence = PulseSequence()
    state1_sequence = PulseSequence()

    RX_pulses = {}
    ro_pulses = {}
    initial_frequency = {}
    for qubit in qubits:
        RX_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX_pulses[qubit].finish
        )
        initial_frequency[qubit] = platform.get_lo_twpa_frequency(qubit)

        state0_sequence.add(ro_pulses[qubit])
        state1_sequence.add(RX_pulses[qubit])
        state1_sequence.add(ro_pulses[qubit])
    sequences = {0: state0_sequence, 1: state1_sequence}
    # create a DataUnits object to store the results
    data = DataUnits(
        name="data",
        quantities={"frequency": "Hz", "delta_frequency": "Hz"},
        options=["qubit", "iteration", "state"],
    )

    # iterate over the frequency range
    delta_frequency_range = np.arange(
        -frequency_width / 2, frequency_width / 2, frequency_step
    ).astype(int)

    # retrieve and store the results for every qubit
    for frequency in delta_frequency_range:
        for qubit in qubits:
            platform.set_lo_twpa_frequency(qubit, initial_frequency[qubit] + frequency)

        # Execute the sequences for both states
        for state in [0, 1]:
            results = platform.execute_pulse_sequence(sequences[state], nshots=nshots)
            for qubit in qubits:
                r = results[ro_pulses[qubit].serial].raw
                r.update(
                    {
                        "frequency[Hz]": [platform.get_lo_twpa_frequency(qubit)]
                        * nshots,
                        "delta_frequency[Hz]": [frequency] * nshots,
                        "qubit": [qubit] * nshots,
                        "iteration": np.arange(nshots),
                        "state": [state] * nshots,
                    }
                )
                data.add_data_from_dict(r)

        # finally, save the remaining data and the fits
        yield data
        yield ro_optimization_fit(
            data, "delta_frequency", "state", "qubit", "iteration"
        )


@plot("TWPA power", plots.ro_power)
def twpa_power(
    platform: AbstractPlatform,
    qubits: dict,
    power_width: float,
    power_step: float,
    nshots: int,
):
    """
    Method which optimizes the Read-out fidelity by varying the power of the TWPA.
    Two analogous tests are performed for calibrate the ground state and the excited state of the oscillator.
    The subscripts `exc` and `gnd` will represent the excited state |1> and the ground state |0>.
    Their distinctiveness is then associated to the fidelity.

    Args:
        platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): custom abstract platform on which we perform the calibration.
        qubits (dict): List of target qubits to perform the action
        power_width (float): width of the power range to be scanned in dBm
        power_step (float): step of the power range to be scanned in dBm
        nshots (int): number of times the pulse sequence will be repeated.
    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **qubit**: The qubit being tested
            - **iteration**: Execution number
            - **state**: The state of the qubit being tested
            - **power**: The power of the TWPA being tested
            - **delta_power**: The power offset from the runcard value

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # create two sequences of pulses:
    # state0_sequence: I  - MZ
    # state1_sequence: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    state0_sequence = PulseSequence()
    state1_sequence = PulseSequence()

    RX_pulses = {}
    ro_pulses = {}
    initial_power = {}
    for qubit in qubits:
        RX_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX_pulses[qubit].finish
        )
        initial_power[qubit] = platform.get_lo_twpa_power(qubit)

        state0_sequence.add(ro_pulses[qubit])
        state1_sequence.add(RX_pulses[qubit])
        state1_sequence.add(ro_pulses[qubit])
    sequences = {0: state0_sequence, 1: state1_sequence}
    # create a DataUnits object to store the results
    data = DataUnits(
        name="data",
        quantities={"power": "dBm", "delta_power": "dBm"},
        options=["qubit", "iteration", "state"],
    )

    # iterate over the power range
    delta_power_range = np.arange(-power_width / 2, power_width / 2, power_step)

    # retrieve and store the results for every qubit
    for power in delta_power_range:
        for qubit in qubits:
            platform.set_lo_twpa_power(qubit, initial_power[qubit] + power)

        # Execute the sequences for both states
        for state in [0, 1]:
            results = platform.execute_pulse_sequence(sequences[state], nshots=nshots)
            for qubit in qubits:
                r = results[ro_pulses[qubit].serial].raw
                r.update(
                    {
                        "power[dBm]": [platform.get_lo_twpa_power(qubit)] * nshots,
                        "delta_power[dBm]": [power] * nshots,
                        "qubit": [qubit] * nshots,
                        "iteration": np.arange(nshots),
                        "state": [state] * nshots,
                    }
                )
                data.add_data_from_dict(r)

        # finally, save the remaining data and the fits
        yield data
        yield ro_optimization_fit(data, "delta_power", "state", "qubit", "iteration")
